from typing import Optional, Set, Union

import numpy as np

from lightautoml.dataset.roles import CategoryRole
from lightautoml.pipelines.selection.base import ImportanceEstimator
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.pipelines.features.base import SparkFeaturesPipeline, SparkTabularDataFeatures
from sparklightautoml.transformers.base import (
    SparkChangeRolesTransformer,
    SparkUnionTransformer,
    SparkSequentialTransformer,
    SparkEstOrTrans,
)

# Same comments as for spark.pipelines.features.base
from sparklightautoml.transformers.categorical import SparkOHEEncoderEstimator, SparkLabelEncoderEstimator
from sparklightautoml.transformers.numeric import (
    SparkFillInfTransformer,
    SparkFillnaMedianEstimator,
    SparkLogOddsTransformer,
    SparkNaNFlagsEstimator,
    SparkStandardScalerEstimator,
)


class SparkLinearFeatures(SparkFeaturesPipeline, SparkTabularDataFeatures):
    """
    Creates pipeline for linear models and nnets.

    Includes:

        - Create categorical intersections.
        - OHE or embed idx encoding for categories.
        - Other cats to numbers ways if defined in role params.
        - Standartization and nan handling for numbers.
        - Numbers discretization if needed.
        - Dates handling.
        - Handling probs (output of lower level models).

    """

    def __init__(
        self,
        feats_imp: Optional[ImportanceEstimator] = None,
        top_intersections: int = 5,
        max_bin_count: int = 10,
        max_intersection_depth: int = 3,
        subsample: Optional[Union[int, float]] = None,
        sparse_ohe: Union[str, bool] = "auto",
        auto_unique_co: int = 50,
        output_categories: bool = True,
        multiclass_te_co: int = 3,
        cacher_key: str = "default_cacher",
        **_
    ):
        """

        Args:
            feats_imp: Features importances mapping.
            top_intersections: Max number of categories
              to generate intersections.
            max_bin_count: Max number of bins to discretize numbers.
            max_intersection_depth: Max depth of cat intersection.
            subsample: Subsample to calc data statistics.
            sparse_ohe: Should we output sparse if ohe encoding
              was used during cat handling.
            auto_unique_co: Switch to target encoding if high cardinality.
            output_categories: Output encoded categories or embed idxs.
            multiclass_te_co: Cutoff if use target encoding in cat handling
              on multiclass task if number of classes is high.

        """
        super().__init__(
            cacher_key=cacher_key,
            multiclass_te_co=multiclass_te_co,
            top_intersections=top_intersections,
            max_intersection_depth=max_intersection_depth,
            subsample=subsample,
            feats_imp=feats_imp,
            auto_unique_co=auto_unique_co,
            output_categories=output_categories,
            ascending_by_cardinality=True,
        )
        # self._input_features = input_features
        # self._input_roles = input_roles

    def _get_input_features(self) -> Set[str]:
        return set(self.input_features)

    def create_pipeline(self, train: SparkDataset) -> SparkEstOrTrans:
        """Create linear pipeline.

        Args:
            train: Dataset with train features.

        Returns:
            Transformer.

        """
        transformers_list = []
        dense_list = []
        sparse_list = []
        probs_list = []
        target_encoder = self.get_target_encoder(train)
        te_list = dense_list if train.task.name == "reg" else probs_list

        # handle categorical feats
        # split categories by handling type. This pipe use 4 encodings - freq/label/target/ohe/ordinal
        # 1 - separate freqs. It does not need label encoding
        dense_list.append(self.get_freq_encoding(train))

        # 2 - check 'auto' type (int is the same - no label encoded numbers in linear models)
        auto = self._cols_by_role(train, "Category", encoding_type="auto") + self._cols_by_role(
            train, "Category", encoding_type="int"
        )

        # if self.output_categories or target_encoder is None:
        if target_encoder is None:
            le = (
                auto
                + self._cols_by_role(train, "Category", encoding_type="oof")
                + self._cols_by_role(train, "Category", encoding_type="ohe")
            )
            te = []

        else:
            te = self._cols_by_role(train, "Category", encoding_type="oof")
            le = self._cols_by_role(train, "Category", encoding_type="ohe")
            # split auto categories by unique values cnt
            un_values = self.get_uniques_cnt(train, auto)
            te = te + [x for x in un_values.index if un_values[x] > self.auto_unique_co]
            le = le + list(set(auto) - set(te))

        # get label encoded categories
        sparse_list.append(self.get_categorical_raw(train, le))

        # get target encoded categories
        te_part = self.get_categorical_raw(train, te)
        if te_part is not None:
            target_encoder_stage = target_encoder(
                input_cols=te_part.getOutputCols(),
                input_roles=te_part.getOutputRoles(),
                task_name=train.task.name,
                folds_column=train.folds_column,
                target_column=train.target_column,
                do_replace_columns=True,
            )
            te_part = SparkSequentialTransformer([te_part, target_encoder_stage])
            te_list.append(te_part)

        # get intersection of top categories
        intersections = self.get_categorical_intersections(train)
        if intersections is not None:
            if target_encoder is not None:
                target_encoder_stage = target_encoder(
                    input_cols=intersections.getOutputCols(),
                    input_roles=intersections.getOutputRoles(),
                    task_name=train.task.name,
                    folds_column=train.folds_column,
                    target_column=train.target_column,
                    do_replace_columns=True,
                )
                ints_part = SparkSequentialTransformer([intersections, target_encoder_stage])
                te_list.append(ints_part)
            else:
                sparse_list.append(intersections)

        # add datetime seasonality
        seas_cats = self.get_datetime_seasons(train, CategoryRole(np.int32))
        if seas_cats is not None:
            # sparse_list.append(SequentialTransformer([seas_cats, LabelEncoder()]))
            label_encoder_stage = SparkLabelEncoderEstimator(
                input_cols=seas_cats.getOutputCols(), input_roles=seas_cats.getOutputRoles(), do_replace_columns=True
            )
            sparse_list.append(SparkSequentialTransformer([seas_cats, label_encoder_stage]))

        # get quantile binning
        sparse_list.append(self.get_binned_data(train))
        # add numeric pipeline wo probs
        dense_list.append(self.get_numeric_data(train, prob=False))
        # add ordinal categories
        dense_list.append(self.get_ordinal_encoding(train))
        # add probs
        probs_list.append(self.get_numeric_data(train, prob=True))
        # add difference with base date
        dense_list.append(self.get_datetime_diffs(train))

        # combine it all together
        # handle probs if exists
        probs_list = [x for x in probs_list if x is not None]
        if len(probs_list) > 0:
            probs_pipe = SparkUnionTransformer(probs_list)
            probs_pipe = SparkSequentialTransformer(
                [
                    probs_pipe,
                    SparkLogOddsTransformer(
                        input_cols=probs_pipe.get_output_cols(),
                        input_roles=probs_pipe.get_output_roles(),
                        do_replace_columns=True,
                    ),
                ]
            )
            dense_list.append(probs_pipe)

        # handle dense
        dense_list = [x for x in dense_list if x is not None]
        if len(dense_list) > 0:
            # standartize, fillna, add null flags

            dense_pipe1 = SparkUnionTransformer(dense_list)
            fill_inf_stage = SparkFillInfTransformer(
                input_cols=dense_pipe1.get_output_cols(),
                input_roles=dense_pipe1.get_output_roles(),
                do_replace_columns=True,
            )
            fill_na_median_stage = SparkFillnaMedianEstimator(
                input_cols=fill_inf_stage.getOutputCols(),
                input_roles=fill_inf_stage.getOutputCols(),
                do_replace_columns=True,
            )
            standerd_scaler_stage = SparkStandardScalerEstimator(
                input_cols=fill_na_median_stage.getOutputCols(),
                input_roles=fill_na_median_stage.getOutputRoles(),
                do_replace_columns=True,
            )

            dense_pipe = SparkSequentialTransformer(
                [
                    dense_pipe1,
                    SparkUnionTransformer(
                        [
                            SparkNaNFlagsEstimator(
                                input_cols=dense_pipe1.get_output_cols(),
                                input_roles=dense_pipe1.get_output_roles(),
                                do_replace_columns=True,
                            ),
                            SparkSequentialTransformer([fill_inf_stage, fill_na_median_stage, standerd_scaler_stage]),
                        ]
                    ),
                ]
            )
            transformers_list.append(dense_pipe)

        # handle categories - cast to float32 if categories are inputs or make ohe
        sparse_list = [x for x in sparse_list if x is not None]
        if len(sparse_list) > 0:
            sparse_pipe = SparkUnionTransformer(sparse_list)
            if self.output_categories:
                final = SparkChangeRolesTransformer(
                    input_cols=sparse_pipe.get_output_cols(),
                    input_roles=sparse_pipe.get_output_roles(),
                    role=CategoryRole(np.float32),
                )
            else:
                assert target_encoder is not None, "Cannot process sparse features cause target encoder is None"
                final = target_encoder(
                    input_cols=sparse_pipe.get_output_cols(),
                    input_roles=sparse_pipe.get_output_roles(),
                    task_name=train.task.name,
                    folds_column=train.folds_column,
                    target_column=train.target_column,
                    do_replace_columns=True,
                )
                # if self.sparse_ohe == "auto":
                #     final = SparkOHEEncoderEstimator(input_cols=sparse_pipe.get_output_cols(),
                #                                      input_roles=sparse_pipe.get_output_roles(),
                #                                      total_feats_cnt=train.shape[1])
                # else:
                #     final = SparkOHEEncoderEstimator(input_cols=sparse_pipe.get_output_cols(),
                #                                      input_roles=sparse_pipe.get_output_roles(),
                #                                      make_sparse=self.sparse_ohe)
            sparse_pipe = SparkSequentialTransformer([sparse_pipe, final])

            transformers_list.append(sparse_pipe)

        # final pipeline
        union_all = SparkUnionTransformer(transformers_list)

        return union_all



from typing import List, Optional

from pyspark.ml.feature import PCA, VectorAssembler
from pyspark.ml.functions import vector_to_array, array_to_vector

from lightautoml.dataset.roles import NumericRole
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.transformers.base import ObsoleteSparkTransformer
from lightautoml.transformers.decomposition import numeric_check

import numpy as np
import pyspark.sql.functions as F


class PCATransformer(ObsoleteSparkTransformer):
    """PCA."""

    _fit_checks = (numeric_check,)
    _transform_checks = ()
    _fname_prefix = "pca"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(self, n_components: int = 500):
        """

        Args:
            n_components: Number of PCA components
        """
        # TODO: should we add support for subs and random_state arguments?

        self.n_components = n_components
        self.pca = None
        self.pca_output_col = "pcaFeatures"
        self._features = None

    def _fit(self, dataset: SparkDataset):
        """Fit algorithm on dataset.

        Args:
            dataset: Sparse or Numpy dataset of text features.

        """

        sdf = dataset.data
        self.n_components = np.minimum(self.n_components, len(sdf.columns) - 1)

        sdf = sdf.select(array_to_vector(F.array("*")).alias("features"))
        pca = PCA(k=3, inputCol="features", outputCol=self.pca_output_col)
        self.pca = pca.fit(sdf)

        orig_name = dataset.features[0].split("__")[-1]

        feats = (
            np.char.array([self._fname_prefix + "_"])
            + np.arange(self.n_components).astype(str)
            + np.char.array(["__" + orig_name])
        )

        self._features = list(feats)
        return self

    def _transform(self, dataset: SparkDataset) -> SparkDataset:
        """Transform input dataset to PCA representation.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            Numpy dataset with text embeddings.

        """

        assert self.pca, "This transformer has not been fitted yet"

        sdf = dataset.data.select(*dataset.service_columns, array_to_vector(F.array("*")).alias("features"))
        new_sdf = self.pca.transform(sdf).select(
            *dataset.service_columns,
            *[
                vector_to_array(F.col(self.pca_output_col))[i].alias(feat)
                for i, feat in zip(range(self.n_components), self.features)
            ]
        )

        output = dataset.empty()
        output.set_data(new_sdf, self.features, NumericRole(np.float32))
        return output



from pyspark import since, keyword_only, SparkContext
from pyspark.ml.param.shared import (
    HasInputCol,
    HasOutputCol,
    HasInputCols,
    HasOutputCols,
    HasHandleInvalid,
    TypeConverters,
    Param,
    Params,
)
from pyspark.ml.util import JavaMLReadable, JavaMLWritable
from pyspark.ml.wrapper import JavaEstimator, JavaModel, JavaParams
from pyspark.ml.common import inherit_doc

from sparklightautoml.mlwriters import LAMLStringIndexerModelJavaMLReadable


class _StringIndexerParams(JavaParams, HasHandleInvalid, HasInputCol, HasOutputCol, HasInputCols, HasOutputCols):
    """
    Params for :py:class:`StringIndexer` and :py:class:`StringIndexerModel`.
    """

    stringOrderType = Param(
        Params._dummy(),
        "stringOrderType",
        "How to order labels of string column. The first label after "
        + "ordering is assigned an index of 0. Supported options: "
        + "frequencyDesc, frequencyAsc, alphabetDesc, alphabetAsc. "
        + "Default is frequencyDesc. In case of equal frequency when "
        + "under frequencyDesc/Asc, the strings are further sorted "
        + "alphabetically",
        typeConverter=TypeConverters.toString,
    )

    handleInvalid = Param(
        Params._dummy(),
        "handleInvalid",
        "how to handle invalid data (unseen "
        + "or NULL values) in features and label column of string type. "
        + "Options are 'skip' (filter out rows with invalid data), "
        + "error (throw an error), or 'keep' (put invalid data "
        + "in a special additional bucket, at index numLabels).",
        typeConverter=TypeConverters.toString,
    )

    minFreqs = Param(
        Params._dummy(),
        "minFreqs",
        "The minimum number of the element occurrences not to skip it.",
        typeConverter=TypeConverters.toListInt,
    )

    defaultValue = Param(
        Params._dummy(),
        "defaultValue",
        "The index for unknown labels and labels that are too rare.",
        typeConverter=TypeConverters.toFloat,
    )

    freqLabel = Param(
        Params._dummy(),
        "freqLabel",
        "If true, label will be transformed to its occurrences",
        typeConverter=TypeConverters.toBoolean,
    )

    nanLast = Param(
        Params._dummy(),
        "nanLast",
        "If true, appends 'NaN' label to the end of each mapping",
        typeConverter=TypeConverters.toBoolean,
    )

    def __init__(self, *args):
        super(_StringIndexerParams, self).__init__(*args)
        self._setDefault(
            handleInvalid="error",
            stringOrderType="frequencyDesc",
            minFreqs=tuple(),
            defaultValue=0.0,
            freqLabel=False,
            nanLast=False,
        )

    @since("2.3.0")
    def getStringOrderType(self):
        """
        Gets the value of :py:attr:`stringOrderType` or its default value 'frequencyDesc'.
        """
        return self.getOrDefault(self.stringOrderType)


class _StringIndexerModelParams(JavaParams, HasHandleInvalid, HasInputCol, HasOutputCol, HasInputCols, HasOutputCols):
    """
    Params for :py:class:`StringIndexer` and :py:class:`StringIndexerModel`.
    """

    # stringOrderType = Param(Params._dummy(), "stringOrderType",
    #                         "How to order labels of string column. The first label after " +
    #                         "ordering is assigned an index of 0. Supported options: " +
    #                         "frequencyDesc, frequencyAsc, alphabetDesc, alphabetAsc. " +
    #                         "Default is frequencyDesc. In case of equal frequency when " +
    #                         "under frequencyDesc/Asc, the strings are further sorted " +
    #                         "alphabetically",
    #                         typeConverter=TypeConverters.toString)

    handleInvalid = Param(
        Params._dummy(),
        "handleInvalid",
        "how to handle invalid data (unseen "
        + "or NULL values) in features and label column of string type. "
        + "Options are 'skip' (filter out rows with invalid data), "
        + "error (throw an error), or 'keep' (put invalid data "
        + "in a special additional bucket, at index numLabels).",
        typeConverter=TypeConverters.toString,
    )

    # minFreqs = Param(Params._dummy(),
    #                  "minFreqs",
    #                  "The minimum number of the element occurrences not to skip it.",
    #                  typeConverter=TypeConverters.toListInt)

    defaultValue = Param(
        Params._dummy(),
        "defaultValue",
        "The index for unknown labels and labels that are too rare.",
        typeConverter=TypeConverters.toFloat,
    )

    freqLabel = Param(
        Params._dummy(),
        "freqLabel",
        "If true, label will be transformed to its occurrences",
        typeConverter=TypeConverters.toBoolean,
    )

    nanLast = Param(
        Params._dummy(),
        "nanLast",
        "If true, appends 'NaN' label to the end of each mapping",
        typeConverter=TypeConverters.toBoolean,
    )

    def __init__(self, *args):
        super(_StringIndexerModelParams, self).__init__(*args)
        self._setDefault(handleInvalid="error", defaultValue=0.0, freqLabel=False, nanLast=False)


@inherit_doc
class LAMLStringIndexer(JavaEstimator, _StringIndexerParams, JavaMLReadable, JavaMLWritable):
    """
    Custom implementation of PySpark StringIndexer wrapper
    """

    @keyword_only
    def __init__(
        self,
        *,
        inputCol=None,
        outputCol=None,
        inputCols=None,
        outputCols=None,
        handleInvalid="error",
        stringOrderType="frequencyDesc",
        minFreqs=None,
        defaultValue=0.0,
        freqLabel=False,
        nanLast=False
    ):
        """
        __init__(self, \\*, inputCol=None, outputCol=None, inputCols=None, outputCols=None, \
                 handleInvalid="error", stringOrderType="frequencyDesc")
        """
        super(LAMLStringIndexer, self).__init__()
        self._java_obj = self._new_java_obj("org.apache.spark.ml.feature.lightautoml.LAMLStringIndexer", self.uid)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    @since("1.4.0")
    def setParams(
        self,
        *,
        inputCol=None,
        outputCol=None,
        inputCols=None,
        outputCols=None,
        handleInvalid="error",
        stringOrderType="frequencyDesc",
        minFreqs=None,
        defaultValue=0.0,
        freqLabel=False,
        nanLast=False
    ):

        """
        setParams(self, \\*, inputCol=None, outputCol=None, inputCols=None, outputCols=None, \
                  handleInvalid="error", stringOrderType="frequencyDesc")
        Sets params for this StringIndexer.
        """

        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def _create_model(self, java_model):
        return LAMLStringIndexerModel(java_model)

    @since("2.3.0")
    def setStringOrderType(self, value):
        """
        Sets the value of :py:attr:`stringOrderType`.
        """
        return self._set(stringOrderType=value)

    def setInputCol(self, value):
        """
        Sets the value of :py:attr:`inputCol`.
        """
        return self._set(inputCol=value)

    @since("3.0.0")
    def setInputCols(self, value):
        """
        Sets the value of :py:attr:`inputCols`.
        """
        return self._set(inputCols=value)

    def setOutputCol(self, value):
        """
        Sets the value of :py:attr:`outputCol`.
        """
        return self._set(outputCol=value)

    @since("3.0.0")
    def setOutputCols(self, value):
        """
        Sets the value of :py:attr:`outputCols`.
        """
        return self._set(outputCols=value)

    def setHandleInvalid(self, value):
        """
        Sets the value of :py:attr:`handleInvalid`.
        """
        return self._set(handleInvalid=value)

    @since("3.2.0")
    def setMinFreqs(self, value):
        """
        Sets the value of :py:attr:`minFreqs`.
        """
        return self._set(minFreqs=value)

    @since("3.2.0")
    def setDefaultValue(self, value):
        """
        Sets the value of :py:attr:`defaultValue`.
        """
        return self._set(defaultValue=value)

    @since("3.2.0")
    def setFreqLabel(self, value):
        """
        Sets the value of :py:attr:`freqLabel`.
        """
        return self._set(freqLabel=value)

    @since("3.2.0")
    def setNanLast(self, value):
        """
        Sets the value of :py:attr:`nanLast`.
        """
        return self._set(nanLast=value)


class LAMLStringIndexerModel(
    JavaModel, _StringIndexerModelParams, LAMLStringIndexerModelJavaMLReadable, JavaMLWritable
):
    """
    Model fitted by :py:class:`StringIndexer`.

    .. versionadded:: 1.4.0
    """

    def setInputCol(self, value):
        """
        Sets the value of :py:attr:`inputCol`.
        """
        return self._set(inputCol=value)

    @since("3.0.0")
    def setInputCols(self, value):
        """
        Sets the value of :py:attr:`inputCols`.
        """
        return self._set(inputCols=value)

    def setOutputCol(self, value):
        """
        Sets the value of :py:attr:`outputCol`.
        """
        return self._set(outputCol=value)

    @since("3.0.0")
    def setOutputCols(self, value):
        """
        Sets the value of :py:attr:`outputCols`.
        """
        return self._set(outputCols=value)

    @since("2.4.0")
    def setHandleInvalid(self, value):
        """
        Sets the value of :py:attr:`handleInvalid`.
        """
        return self._set(handleInvalid=value)

    @since("3.2.0")
    def setDefaultValue(self, value):
        """
        Sets the value of :py:attr:`defaultValue`.
        """
        return self._set(defaultValue=value)

    @since("3.2.0")
    def setFreqLabel(self, value):
        """
        Sets the value of :py:attr:`freqLabel`.
        """
        return self._set(freqLabel=value)

    @since("3.2.0")
    def setNanLast(self, value):
        """
        Sets the value of :py:attr:`nanLast`.
        """
        return self._set(nanLast=value)

    @classmethod
    @since("2.4.0")
    def from_labels(
        cls, labels, inputCol, outputCol=None, handleInvalid=None, defaultValue=0.0, freqLabel=False, nanLast=False
    ):
        """
        Construct the model directly from an array of label strings,
        requires an active SparkContext.
        """
        sc = SparkContext._active_spark_context
        java_class = sc._gateway.jvm.java.lang.String
        jlabels = LAMLStringIndexerModel._new_java_array(labels, java_class)
        model = LAMLStringIndexerModel._create_from_java_class(
            "org.apache.spark.ml.feature.lightautoml.LAMLStringIndexerModel", jlabels
        )
        model.setInputCol(inputCol)
        if outputCol is not None:
            model.setOutputCol(outputCol)
        if handleInvalid is not None:
            model.setHandleInvalid(handleInvalid)

        model.setDefaultValue(defaultValue)
        model.setFreqLabel(freqLabel)
        model.setNanLast(nanLast)

        return model

    @classmethod
    @since("3.0.0")
    def from_arrays_of_labels(
        cls, arrayOfLabels, inputCols, outputCols=None, handleInvalid=None, defaultValue=0.0, freqLabel=False
    ):
        """
        Construct the model directly from an array of array of label strings,
        requires an active SparkContext.
        """
        sc = SparkContext._active_spark_context
        java_class = sc._gateway.jvm.java.lang.String
        jlabels = LAMLStringIndexerModel._new_java_array(arrayOfLabels, java_class)
        model = LAMLStringIndexerModel._create_from_java_class(
            "org.apache.spark.ml.feature.lightautoml.LAMLStringIndexerModel", jlabels
        )
        model.setInputCols(inputCols)
        if outputCols is not None:
            model.setOutputCols(outputCols)
        if handleInvalid is not None:
            model.setHandleInvalid(handleInvalid)

        model.setDefaultValue(defaultValue)
        model.setFreqLabel(freqLabel)

        return model

    @property
    @since("1.5.0")
    def labels(self):
        """
        Ordered list of labels, corresponding to indices to be assigned.

        .. deprecated:: 3.1.0
            It will be removed in future versions. Use `labelsArray` method instead.
        """
        return self._call_java("labels")

    @property
    @since("3.0.2")
    def labelsArray(self):
        """
        Array of ordered list of labels, corresponding to indices to be assigned
        for each input column.
        """
        return self._call_java("getStringLabels")

    @staticmethod
    def _from_java(java_stage):
        """
        Given a Java object, create and return a Python wrapper of it.
        Used for ML persistence.

        Meta-algorithms such as Pipeline should override this method as a classmethod.
        """

        def __get_class(clazz):
            """
            Loads Python class from its name.
            """
            parts = clazz.split(".")
            module = ".".join(parts[:-1])
            m = __import__(module)
            for comp in parts[1:]:
                m = getattr(m, comp)
            return m

        stage_name = "sparklightautoml.transformers.scala_wrappers.laml_string_indexer.LAMLStringIndexerModel"
        # Generate a default new instance from the stage_name class.
        py_type = __get_class(stage_name)
        if issubclass(py_type, JavaParams):
            # Load information from java_stage to the instance.
            py_stage = py_type()
            py_stage._java_obj = java_stage
            py_stage._resetUid(java_stage.uid())
            py_stage._transfer_params_from_java()
        elif hasattr(py_type, "_from_java"):
            py_stage = py_type._from_java(java_stage)
        else:
            raise NotImplementedError("This Java stage cannot be loaded into Python currently: %r" % stage_name)
        return py_stage



import itertools
import logging
from collections import defaultdict
from dataclasses import dataclass
from itertools import combinations
from typing import Optional, Sequence, List, Tuple, Dict, Any

import numpy as np
import pandas as pd
from pandas import Series
from pyspark.ml import Transformer
from pyspark.ml.feature import OneHotEncoder
from pyspark.ml.param.shared import Param, Params
from pyspark.sql import functions as F, types as SparkTypes, DataFrame as SparkDataFrame, Window, Column
from sklearn.utils.murmurhash import murmurhash3_32

from lightautoml.dataset.base import RolesDict
from lightautoml.dataset.roles import CategoryRole, NumericRole, ColumnRole
from sparklightautoml.dataset.roles import NumericVectorOrArrayRole
from sparklightautoml.mlwriters import (
    CommonPickleMLReadable,
    CommonPickleMLWritable,
    SparkLabelEncoderTransformerMLReadable,
    SparkLabelEncoderTransformerMLWritable,
)
from sparklightautoml.transformers.base import SparkBaseEstimator, SparkBaseTransformer
from lightautoml.transformers.categorical import (
    categorical_check,
    encoding_check,
    oof_task_check,
    multiclass_task_check,
)

from sparklightautoml.transformers.scala_wrappers.laml_string_indexer import LAMLStringIndexer, LAMLStringIndexerModel

logger = logging.getLogger(__name__)


# FIXME SPARK-LAMA: np.nan in str representation is 'nan' while Spark's NaN is 'NaN'. It leads to different hashes.
# FIXME SPARK-LAMA: If udf is defined inside the class, it not works properly.
# "if murmurhash3_32 can be applied to a whole pandas Series, it would be better to make it via pandas_udf"
# https://github.com/fonhorst/LightAutoML/pull/57/files/57c15690d66fbd96f3ee838500de96c4637d59fe#r749534669
murmurhash3_32_udf = F.udf(
    lambda value: murmurhash3_32(value.replace("NaN", "nan"), seed=42) if value is not None else None,
    SparkTypes.IntegerType(),
)


@dataclass
class OOfFeatsMapping:
    folds_column: str
    # count of different categories in the categorical column being processed
    dim_size: int
    # mapping from category to a continious reperesentation found by target encoder
    # category may be represented:
    # - cat (plain category)
    # - dim_size * folds_num + cat
    # mapping: Dict[int, float]
    mapping: np.array


class TypesHelper:
    _ad_hoc_types_mapper = defaultdict(
        lambda: "string",
        {
            "bool": "boolean",
            "int": "int",
            "int8": "int",
            "int16": "int",
            "int32": "int",
            "int64": "int",
            "int128": "bigint",
            "int256": "bigint",
            "integer": "int",
            "uint8": "int",
            "uint16": "int",
            "uint32": "int",
            "uint64": "int",
            "uint128": "bigint",
            "uint256": "bigint",
            "longlong": "long",
            "ulonglong": "long",
            "float16": "float",
            "float": "float",
            "float32": "float",
            "float64": "double",
            "float128": "double",
        },
    )

    _spark_numeric_types_str = ("ShortType", "IntegerType", "LongType", "FloatType", "DoubleType", "DecimalType")


def pandas_dict_udf(broadcasted_dict):
    def f(s: Series) -> Series:
        values_dict = broadcasted_dict.value
        return s.map(values_dict)

    return F.pandas_udf(f, "double")


def pandas_1d_mapping_udf(broadcasted_arr):
    def f(s: Series) -> Series:
        s_na = s.isna()
        s_fna = s.fillna(0).astype("int32")
        np_arr = broadcasted_arr.value
        res = pd.Series(np_arr[s_fna])
        res[s_na] = np.nan
        return res

    return F.pandas_udf(f, "double")


def pandas_folds_cat_mapping_udf(broadcasted_arr):
    def f(folds: Series, cat: Series) -> Series:
        np_arr = broadcasted_arr.value
        return pd.Series(np_arr[folds, cat])

    return F.pandas_udf(f, "double")


class SparkLabelEncoderEstimator(SparkBaseEstimator, TypesHelper):
    """
    Spark label encoder estimator. Returns :class:`~sparklightautoml.transformers.categorical.SparkLabelEncoderTransformer`.
    """

    _fit_checks = (categorical_check,)
    _transform_checks = ()
    _fname_prefix = "le"

    _fillna_val = 0.0

    def __init__(
        self,
        input_cols: Optional[List[str]] = None,
        input_roles: Optional[Dict[str, ColumnRole]] = None,
        subs: Optional[float] = None,
        random_state: Optional[int] = 42,
        do_replace_columns: bool = False,
        output_role: Optional[ColumnRole] = None,
    ):
        if not output_role:
            output_role = CategoryRole(np.int32, label_encoded=True)
        super().__init__(input_cols, input_roles, do_replace_columns=do_replace_columns, output_role=output_role)
        self._input_intermediate_columns = self.getInputCols()
        self._input_intermediate_roles = self.getInputRoles()

    def _fit(self, dataset: SparkDataFrame) -> "SparkLabelEncoderTransformer":
        logger.info(f"[{type(self)} (LE)] fit is started")

        roles = self._input_intermediate_roles
        columns = self._input_intermediate_columns

        if self._fname_prefix == "inter":
            roles = self._input_roles
            columns = self._input_columns

        indexer = LAMLStringIndexer(
            inputCols=columns,
            outputCols=self.getOutputCols(),
            minFreqs=[roles[col_name].unknown for col_name in columns],
            handleInvalid="keep",
            defaultValue=self._fillna_val,
        )

        self.indexer_model = indexer.fit(dataset)

        logger.info(f"[{type(self)} (LE)] fit is finished")

        return SparkLabelEncoderTransformer(
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.getInputRoles(),
            output_roles=self.getOutputRoles(),
            do_replace_columns=self.getDoReplaceColumns(),
            indexer_model=self.indexer_model,
        )


class SparkLabelEncoderTransformer(
    SparkBaseTransformer, TypesHelper, SparkLabelEncoderTransformerMLWritable, SparkLabelEncoderTransformerMLReadable
):
    """
    Simple Spark version of `LabelEncoder`.

    Labels are integers from 1 to n.
    """

    _transform_checks = ()
    _fname_prefix = "le"

    _fillna_val = 0.0

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool,
        indexer_model: LAMLStringIndexerModel,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)
        self.indexer_model = indexer_model

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:

        logger.info(f"[{type(self)} (LE)] transform is started")
        columns = self.getInputCols()
        out_columns = self.getOutputCols()
        if self._fname_prefix == "inter":
            columns = self._input_columns
            out_columns = sorted(out_columns)

        model: LAMLStringIndexerModel = (
            self.indexer_model.setDefaultValue(self._fillna_val)
            .setHandleInvalid("keep")
            .setInputCols(columns)
            .setOutputCols(out_columns)
        )

        logger.info(f"[{type(self)} (LE)] Transform is finished")

        # output = self._make_output_df(model.transform(dataset), self.getOutputCols())
        output = model.transform(dataset)

        return output


class SparkOrdinalEncoderEstimator(SparkLabelEncoderEstimator):
    """
    Spark ordinal encoder estimator. Returns :class:`~sparklightautoml.transformers.categorical.SparkOrdinalEncoderTransformer`.
    """

    _fit_checks = (categorical_check,)
    _fname_prefix = "ord"
    _fillna_val = float("nan")

    def __init__(
        self,
        input_cols: Optional[List[str]] = None,
        input_roles: Optional[Dict[str, ColumnRole]] = None,
        subs: Optional[float] = None,
        random_state: Optional[int] = 42,
    ):
        super().__init__(input_cols, input_roles, subs, random_state, output_role=NumericRole(np.float32))
        self.dicts = None
        self._use_cols = self.getInputCols()

    def _fit(self, dataset: SparkDataFrame) -> "Transformer":

        logger.info(f"[{type(self)} (ORD)] fit is started")

        cols_to_process = [
            col for col in self.getInputCols() if str(dataset.schema[col].dataType) not in self._spark_numeric_types_str
        ]

        min_freqs = [self._input_intermediate_roles[col].unknown for col in cols_to_process]

        indexer = LAMLStringIndexer(
            stringOrderType="alphabetAsc",
            inputCols=cols_to_process,
            outputCols=[f"{self._fname_prefix}__{col}" for col in cols_to_process],
            minFreqs=min_freqs,
            handleInvalid="keep",
            defaultValue=self._fillna_val,
            nanLast=True,  # Only for ORD
        )

        self.indexer_model = indexer.fit(dataset)

        logger.info(f"[{type(self)} (ORD)] fit is finished")

        return SparkOrdinalEncoderTransformer(
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.getInputRoles(),
            output_roles=self.getOutputRoles(),
            do_replace_columns=self.getDoReplaceColumns(),
            indexer_model=self.indexer_model,
        )


class SparkOrdinalEncoderTransformer(SparkLabelEncoderTransformer):
    """
    Spark version of :class:`~lightautoml.transformers.categorical.OrdinalEncoder`.

    Encoding ordinal categories into numbers.
    Number type categories passed as is,
    object type sorted in ascending lexicographical order.
    """

    _transform_checks = ()
    _fname_prefix = "ord"
    _fillna_val = float("nan")

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool,
        indexer_model: LAMLStringIndexerModel,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns, indexer_model)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:

        logger.info(f"[{type(self)} (ORD)] transform is started")

        input_columns = self.getInputCols()
        output_columns = self.getOutputCols()

        cols_to_process = [
            col for col in self.getInputCols() if str(dataset.schema[col].dataType) not in self._spark_numeric_types_str
        ]

        # TODO: Fix naming
        transformed_cols = [f"{self._fname_prefix}__{col}" for col in cols_to_process]

        model: LAMLStringIndexerModel = (
            self.indexer_model.setDefaultValue(self._fillna_val)
            .setHandleInvalid("keep")
            .setInputCols(cols_to_process)
            .setOutputCols(transformed_cols)
        )

        indexed_dataset = model.transform(dataset)

        for input_col, output_col in zip(input_columns, output_columns):
            if output_col in transformed_cols:
                continue
            indexed_dataset = indexed_dataset.withColumn(output_col, F.col(input_col))

        indexed_dataset = indexed_dataset.replace(float("nan"), 0.0, subset=output_columns)

        logger.info(f"[{type(self)} (ORD)] Transform is finished")

        # output = self._make_output_df(indexed_dataset, self.getOutputCols())
        output = indexed_dataset

        return output


class SparkFreqEncoderEstimator(SparkLabelEncoderEstimator):
    """
    Calculates frequency in train data and produces :class:`~sparklightautoml.transformers.categorical.SparkFreqEncoderTransformer` instance.
    """

    _fit_checks = (categorical_check,)
    _transform_checks = ()
    _fname_prefix = "freq"

    _fillna_val = 1

    def __init__(self, input_cols: List[str], input_roles: RolesDict, do_replace_columns: bool = False):
        super().__init__(input_cols, input_roles, do_replace_columns, output_role=NumericRole(np.float32))

    def _fit(self, dataset: SparkDataFrame) -> "SparkFreqEncoderTransformer":

        logger.info(f"[{type(self)} (FE)] fit is started")

        indexer = LAMLStringIndexer(
            inputCols=self._input_intermediate_columns,
            outputCols=self.getOutputCols(),
            minFreqs=[1 for _ in self._input_intermediate_columns],
            handleInvalid="keep",
            defaultValue=self._fillna_val,
            freqLabel=True,  # Only for FREQ encoder
        )

        self.indexer_model = indexer.fit(dataset)

        logger.info(f"[{type(self)} (FE)] fit is finished")

        return SparkFreqEncoderTransformer(
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.getInputRoles(),
            output_roles=self.getOutputRoles(),
            do_replace_columns=self.getDoReplaceColumns(),
            indexer_model=self.indexer_model,
        )


class SparkFreqEncoderTransformer(SparkLabelEncoderTransformer):
    """
    Labels are encoded with frequency in train data.

    Labels are integers from 1 to n.
    """

    _fit_checks = (categorical_check,)
    _transform_checks = ()
    _fname_prefix = "freq"
    _fillna_val = 1.0

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool,
        indexer_model: LAMLStringIndexerModel,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns, indexer_model)


class SparkCatIntersectionsHelper:
    """Helper class for :class:`~sparklightautoml.transformers.categorical.SparkCatIntersectionsEstimator` and
    :class:`~sparklightautoml.transformers.categorical.SparkCatIntersectionsTransformer`.
    """

    _fname_prefix = "inter"

    def _make_col_name(self, cols: Sequence[str]) -> str:
        return f"({'__'.join(cols)})"

    def _make_category(self, cols: Sequence[str]) -> Column:
        lit = F.lit("_")
        col_name = self._make_col_name(cols)
        columns_for_concat = []
        for col in cols:
            columns_for_concat.append(F.col(col))
            columns_for_concat.append(lit)
        columns_for_concat = columns_for_concat[:-1]

        return murmurhash3_32_udf(F.concat(*columns_for_concat)).alias(col_name)

    def _build_df(self, df: SparkDataFrame, intersections: Optional[Sequence[Sequence[str]]]) -> SparkDataFrame:
        columns_to_select = [self._make_category(comb).alias(f"{self._make_col_name(comb)}") for comb in intersections]
        df = df.select("*", *columns_to_select)
        return df


class SparkCatIntersectionsEstimator(SparkCatIntersectionsHelper, SparkLabelEncoderEstimator):
    """
    Combines categorical features and fits :class:`~sparklightautoml.transformers.categorical.SparkLabelEncoderEstimator`.
    Returns :class:`~sparklightautoml.transformers.categorical.SparkCatIntersectionsTransformer`.
    """

    _fit_checks = (categorical_check,)
    _transform_checks = ()
    _fname_prefix = "inter"

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        intersections: Optional[Sequence[Sequence[str]]] = None,
        max_depth: int = 2,
        do_replace_columns: bool = False,
    ):

        super().__init__(
            input_cols,
            input_roles,
            do_replace_columns=do_replace_columns,
            output_role=CategoryRole(np.int32, label_encoded=True),
        )
        self.intersections = intersections
        self.max_depth = max_depth

        if self.intersections is None:
            self.intersections = []
            for i in range(2, min(self.max_depth, len(self.getInputCols())) + 1):
                self.intersections.extend(list(combinations(self.getInputCols(), i)))

        self._input_roles = {
            f"{self._make_col_name(comb)}": CategoryRole(
                np.int32,
                unknown=max((self.getInputRoles()[x].unknown for x in comb)),
                label_encoded=True,
            )
            for comb in self.intersections
        }
        self._input_columns = sorted(list(self._input_roles.keys()))

        out_roles = {f"{self._fname_prefix}__{f}": role for f, role in self._input_roles.items()}

        self.set(self.outputCols, list(out_roles.keys()))
        self.set(self.outputRoles, out_roles)

    def _fit(self, df: SparkDataFrame) -> Transformer:
        logger.info(f"[{type(self)} (CI)] fit is started")
        logger.debug(f"Calculating (CI) for input columns: {self.getInputCols()}")

        inter_df = self._build_df(df, self.intersections)

        super()._fit(inter_df)

        logger.info(f"[{type(self)} (CI)] fit is finished")

        return SparkCatIntersectionsTransformer(
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.getInputRoles(),
            output_roles=self.getOutputRoles(),
            do_replace_columns=self.getDoReplaceColumns(),
            indexer_model=self.indexer_model,
            intersections=self.intersections,
        )


class SparkCatIntersectionsTransformer(SparkCatIntersectionsHelper, SparkLabelEncoderTransformer):
    """
    Combines category columns and encode with label encoder.
    """

    _fit_checks = (categorical_check,)
    _transform_checks = ()

    _fillna_val = 0

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool,
        indexer_model: LAMLStringIndexerModel,
        intersections: Optional[Sequence[Sequence[str]]] = None,
    ):

        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns, indexer_model)
        self.intersections = intersections

    def _transform(self, df: SparkDataFrame) -> SparkDataFrame:
        inter_df = self._build_df(df, self.intersections)
        temp_cols = sorted(list(set(inter_df.columns).difference(df.columns)))
        self._input_columns = temp_cols

        out_df = super()._transform(inter_df)

        out_df = out_df.drop(*temp_cols)
        return out_df


class SparkOHEEncoderEstimator(SparkBaseEstimator):
    """
    Simple OneHotEncoder over label encoded categories.
    """

    _fit_checks = (categorical_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "ohe"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        do_replace_columns: bool = False,
        make_sparse: Optional[bool] = None,
        total_feats_cnt: Optional[int] = None,
        dtype: type = np.float32,
    ):
        """
        Args:
            make_sparse: Create sparse matrix.
            total_feats_cnt: Initial features number.
            dtype: Dtype of new features.
        """
        super().__init__(input_cols, input_roles, do_replace_columns=do_replace_columns, output_role=None)

        self._make_sparse = make_sparse
        self._total_feats_cnt = total_feats_cnt
        self.dtype = dtype

        if self._make_sparse is None:
            assert self._total_feats_cnt is not None, "Param total_feats_cnt should be defined if make_sparse is None"

        self._ohe_transformer_and_roles: Optional[Tuple[Transformer, Dict[str, ColumnRole]]] = None

    def _fit(self, sdf: SparkDataFrame) -> Transformer:
        """Calc output shapes.
        Automatically do ohe in sparse form if approximate fill_rate < `0.2`.
        Args:
            dataset: Pandas or Numpy dataset of categorical features.
        Returns:
            self.
        """

        maxs = [F.max(c).alias(f"max_{c}") for c in self.getInputCols()]
        mins = [F.min(c).alias(f"min_{c}") for c in self.getInputCols()]
        mm = sdf.select(maxs + mins).first().asDict()

        ohe = OneHotEncoder(inputCols=self.getInputCols(), outputCols=self.getOutputCols(), handleInvalid="error")
        transformer = ohe.fit(sdf)

        roles = {
            f"{self._fname_prefix}__{c}": NumericVectorOrArrayRole(
                size=mm[f"max_{c}"] - mm[f"min_{c}"] + 1,
                element_col_name_template=[
                    f"{self._fname_prefix}_{i}__{c}" for i in np.arange(mm[f"min_{c}"], mm[f"max_{c}"] + 1)
                ],
            )
            for c in self.getInputCols()
        }

        self._ohe_transformer_and_roles = (transformer, roles)

        return OHEEncoderTransformer(
            transformer,
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.getInputRoles(),
            output_roles=roles,
        )


class OHEEncoderTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """OHEEncoder Transformer"""

    _fit_checks = (categorical_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "ohe"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self._features

    def __init__(
        self,
        ohe_transformer: Transformer,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool = False,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)
        self._ohe_transformer = ohe_transformer

    def _transform(self, sdf: SparkDataFrame) -> SparkDataFrame:
        """Transform categorical dataset to ohe.
        Args:
            dataset: Pandas or Numpy dataset of categorical features.
        Returns:
            Numpy dataset with encoded labels.
        """
        output = self._ohe_transformer.transform(sdf)

        return output


def te_mapping_udf(broadcasted_dict):
    def f(folds, current_column):
        values_dict = broadcasted_dict.value
        try:
            return values_dict[f"{folds}_{current_column}"]
        except KeyError:
            return np.nan

    return F.udf(f, "double")


class SparkTargetEncoderEstimator(SparkBaseEstimator):
    """
    Spark target encoder estimator. Returns :class:`~sparklightautoml.transformers.categorical.SparkTargetEncoderTransformer`.
    """

    _fit_checks = (categorical_check, oof_task_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "oof"

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        alphas: Sequence[float] = (0.5, 1.0, 2.0, 5.0, 10.0, 50.0, 250.0, 1000.0),
        task_name: Optional[str] = None,
        folds_column: Optional[str] = None,
        target_column: Optional[str] = None,
        do_replace_columns: bool = False,
    ):
        super().__init__(
            input_cols, input_roles, do_replace_columns, NumericRole(np.float32, prob=task_name == "binary")
        )
        self.alphas = alphas
        self._task_name = task_name
        self._folds_column = folds_column
        self._target_column = target_column

    @staticmethod
    def score_func_binary(target, candidate) -> float:
        return -(target * np.log(candidate) + (1 - target) * np.log(1 - candidate))

    @staticmethod
    def score_func_reg(target, candidate) -> float:
        return (target - candidate) ** 2

    def _fit(self, dataset: SparkDataFrame) -> "SparkTargetEncoderTransformer":
        logger.info(f"[{type(self)} (TE)] fit_transform is started")

        assert self._target_column in dataset.columns, "Target should be presented in the dataframe"
        assert self._folds_column in dataset.columns, "Folds should be presented in the dataframe"

        self.encodings: Dict[str, Dict[int, float]] = dict()
        oof_feats_encoding: Dict[str, OOfFeatsMapping] = dict()

        sdf = dataset

        _fc = F.col(self._folds_column)
        _tc = F.col(self._target_column)

        logger.debug("Calculating totals (TE)")
        n_folds, prior, total_target_sum, total_count = sdf.select(
            F.max(_fc) + 1, F.mean(_tc.cast("double")), F.sum(_tc).cast("double"), F.count(_tc)
        ).first()

        logger.debug("Calculating folds priors (TE)")
        folds_prior_pdf = (
            sdf.groupBy(_fc)
            .agg(((total_target_sum - F.sum(_tc)) / (total_count - F.count(_tc))).alias("_folds_prior"))
            .collect()
        )

        def binary_score(col_name: str):
            return F.mean(-(_tc * F.log(col_name) + (1 - _tc) * F.log(1 - F.col(col_name)))).alias(col_name)

        def reg_score(col_name: str):
            return F.mean(F.pow((_tc - F.col(col_name)), F.lit(2))).alias(col_name)

        logger.debug("Starting processing features")
        feature_count = len(self.getInputCols())
        for i, feature in enumerate(self.getInputCols()):
            logger.debug(f"Processing feature {feature}({i}/{feature_count})")

            _cur_col = F.col(feature)
            (dim_size,) = sdf.select((F.max(_cur_col) + 1).astype("int").alias("dim_size")).first()

            logger.debug(f"Dim size of feature {feature}: {dim_size}")

            windowSpec = Window.partitionBy(_cur_col)
            f_df = sdf.groupBy(_cur_col, _fc).agg(F.sum(_tc).alias("f_sum"), F.count(_tc).alias("f_count")).cache()

            oof_df = f_df.select(
                _cur_col,
                _fc,
                (F.sum("f_sum").over(windowSpec) - F.col("f_sum")).alias("oof_sum"),
                (F.sum("f_count").over(windowSpec) - F.col("f_count")).alias("oof_count"),
            )

            logger.debug(f"Creating maps column for fold priors (size={len(folds_prior_pdf)}) (TE)")
            mapping = {row[self._folds_column]: row["_folds_prior"] for row in folds_prior_pdf}
            folds_prior_exp = F.create_map(*[F.lit(x) for x in itertools.chain(*mapping.items())])

            logger.debug(f"Creating candidate columns (count={len(self.alphas)}) (TE)")
            candidates_cols = [
                ((F.col("oof_sum") + F.lit(alpha) * folds_prior_exp[_fc]) / (F.col("oof_count") + F.lit(alpha)))
                .cast("double")
                .alias(f"candidate_{i}")
                for i, alpha in enumerate(self.alphas)
            ]

            candidates_df = oof_df.select(_cur_col, _fc, *candidates_cols).cache()

            score_func = binary_score if self._task_name == "binary" else reg_score

            logger.debug("Calculating scores (TE)")
            scores = (
                sdf.join(candidates_df, on=[feature, self._folds_column])
                .select(*[score_func(f"candidate_{i}") for i, alpha in enumerate(self.alphas)])
                .first()
                .asDict()
            )
            logger.debug(f"Scores have been calculated (size={len(scores)}) (TE)")

            seq_scores = [scores[f"candidate_{i}"] for i, alpha in enumerate(self.alphas)]
            best_alpha_idx = np.argmin(seq_scores)
            best_alpha = self.alphas[best_alpha_idx]

            logger.debug("Collecting encodings (TE)")
            encoding_df = f_df.groupby(_cur_col).agg(
                ((F.sum("f_sum") + best_alpha * prior) / (F.sum("f_count") + best_alpha)).alias("encoding")
            )
            encoding = encoding_df.toPandas()
            logger.debug(f"Encodings have been collected (size={len(encoding)}) (TE)")
            f_df.unpersist()

            mapping = np.zeros(dim_size, dtype=np.float64)
            np.add.at(mapping, encoding[feature].astype(np.int32).to_numpy(), encoding["encoding"])
            self.encodings[feature] = mapping

            logger.debug("Collecting oof_feats (TE)")
            oof_feats_df = candidates_df.select(_cur_col, _fc, F.col(f"candidate_{best_alpha_idx}").alias("encoding"))
            oof_feats = oof_feats_df.toPandas()
            logger.debug(f"oof_feats have been collected (size={len(oof_feats)}) (TE)")

            candidates_df.unpersist()

            mapping = np.zeros((n_folds, dim_size), dtype=np.float64)
            np.add.at(
                mapping,
                (
                    oof_feats[self._folds_column].astype(np.int32).to_numpy(),
                    oof_feats[feature].astype(np.int32).to_numpy(),
                ),
                oof_feats["encoding"],
            )

            oof_feats = OOfFeatsMapping(folds_column=self._folds_column, dim_size=dim_size, mapping=mapping)

            oof_feats_encoding[feature] = oof_feats

            logger.debug(f"[{type(self)} (TE)] Encodings have been calculated")

        logger.info(f"[{type(self)} (TE)] fit_transform is finished")

        return SparkTargetEncoderTransformer(
            encodings=self.encodings,
            input_cols=self.getInputCols(),
            input_roles=self.getInputRoles(),
            output_cols=self.getOutputCols(),
            output_roles=self.getOutputRoles(),
            do_replace_columns=self.getDoReplaceColumns(),
            oof_feats_encoding=oof_feats_encoding,
        )


class SparkTargetEncoderTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """
    Spark version of :class:`~lightautoml.transformers.categorical.TargetEncoder`.
    """

    _fit_checks = (categorical_check, oof_task_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "oof"

    def __init__(
        self,
        encodings: Dict[str, np.array],
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool = False,
        oof_feats_encoding: Optional[Dict[str, OOfFeatsMapping]] = None,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)
        self._encodings = encodings
        self._oof_feats_encoding = oof_feats_encoding

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:

        cols_to_select = []
        logger.info(f"[{type(self)} (TE)] transform is started")

        sc = dataset.sql_ctx.sparkSession.sparkContext

        for col_name, out_name in zip(self.getInputCols(), self.getOutputCols()):
            _cur_col = F.col(col_name)

            if self._oof_feats_encoding is not None:
                oof_feats: OOfFeatsMapping = self._oof_feats_encoding[col_name]
                _fc = F.col(oof_feats.folds_column)
                # this is necessary to process data differently first time
                # e.g. during pipeline fitting
                values = sc.broadcast(oof_feats.mapping)
                cols_to_select.append(
                    pandas_folds_cat_mapping_udf(values)(_fc.astype("int"), _cur_col.astype("int")).alias(out_name)
                )
            else:
                logger.debug(
                    f"[{type(self)} (TE)] transform map size for column {col_name}: {len(self._encodings[col_name])}"
                )
                values = sc.broadcast(self._encodings[col_name])
                cols_to_select.append(pandas_1d_mapping_udf(values)(_cur_col).alias(out_name))

        output = self._make_output_df(dataset, cols_to_select)

        out_cols = [
            F.when(F.isnull(out_col), float("nan")).otherwise(F.col(out_col)).alias(out_col)
            for out_col in self.getOutputCols()
        ]
        rest_cols = [c for c in output.columns if c not in self.getOutputCols()]
        output = output.select(*rest_cols, *out_cols)

        self._oof_feats_encoding = None

        logger.info(f"[{type(self)} (TE)] transform is finished")

        return output


# def mcte_mapping_udf(broadcasted_dict):
#     def f(folds, target, current_column):
#         values_dict = broadcasted_dict.value
#         try:
#             return values_dict[(folds, target, current_column)]
#         except KeyError:
#             return np.nan
#     return F.udf(f, "double")


def mcte_transform_udf(broadcasted_dict):
    def f(target, current_column):
        values_dict = broadcasted_dict.value
        try:
            return values_dict[(target, current_column)]
        except KeyError:
            return np.nan

    return F.udf(f, "double")


class SparkMulticlassTargetEncoderEstimator(SparkBaseEstimator):
    """
    Spark multiclass target encoder estimator. Returns :class:`~sparklightautoml.transformers.categorical.SparkMultiTargetEncoderTransformer`.
    """

    _fit_checks = (categorical_check, multiclass_task_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "multioof"

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        alphas: Sequence[float] = (0.5, 1.0, 2.0, 5.0, 10.0, 50.0, 250.0, 1000.0),
        task_name: Optional[str] = None,
        folds_column: Optional[str] = None,
        target_column: Optional[str] = None,
        do_replace_columns: bool = False,
    ):
        super().__init__(input_cols, input_roles, do_replace_columns, NumericRole(np.float32, prob=True))
        self.alphas = alphas
        self._task_name = task_name
        self._folds_column = folds_column
        self._target_column = target_column

    def _fit(self, dataset: SparkDataFrame) -> "SparkMultiTargetEncoderTransformer":
        logger.info(f"[{type(self)} (MCTE)] fit_transform is started")

        assert self._target_column in dataset.columns, "Target should be presented in the dataframe"
        assert self._folds_column in dataset.columns, "Folds should be presented in the dataframe"

        self.encodings = []

        df = dataset
        sc = df.sql_ctx.sparkSession.sparkContext

        _fc = F.col(self._folds_column)
        _tc = F.col(self._target_column)

        tcn = self._target_column
        fcn = self._folds_column

        agg = df.groupBy([_fc, _tc]).count().toPandas().sort_values(by=[fcn, tcn])

        rows_count = agg["count"].sum()
        prior = agg.groupby(tcn).agg({"count": sum})

        prior["prior"] = prior["count"] / float(rows_count)
        prior = prior.to_dict()["prior"]

        agg["tt_sum"] = agg[tcn].map(agg[[tcn, "count"]].groupby(tcn).sum()["count"].to_dict()) - agg["count"]
        agg["tf_sum"] = rows_count - agg[fcn].map(agg[[fcn, "count"]].groupby(fcn).sum()["count"].to_dict())

        agg["folds_prior"] = agg["tt_sum"] / agg["tf_sum"]
        folds_prior_dict = agg[[fcn, tcn, "folds_prior"]].groupby([fcn, tcn]).max().to_dict()["folds_prior"]

        # Folds column unique values
        fcvs = sorted(list(set([fold for fold, target in folds_prior_dict.keys()])))
        # Target column unique values
        tcvs = sorted(list(set([target for fold, target in folds_prior_dict.keys()])))

        # cols_to_select = []

        for ccn in self.getInputCols():

            logger.debug(f"[{type(self)} (MCTE)] column {ccn}")

            _cc = F.col(ccn)

            col_agg = df.groupby(_fc, _tc, _cc).count().toPandas()
            col_agg_dict = col_agg.groupby([ccn, fcn, tcn]).sum().to_dict()["count"]
            t_sum_dict = col_agg[[ccn, tcn, "count"]].groupby([ccn, tcn]).sum().to_dict()["count"]
            f_count_dict = col_agg[[ccn, fcn, "count"]].groupby([ccn, fcn]).sum().to_dict()["count"]
            t_count_dict = col_agg[[ccn, "count"]].groupby([ccn]).sum().to_dict()["count"]

            alphas_values = dict()
            # Current column unique values
            ccvs = sorted(col_agg[ccn].unique())

            for column_value in ccvs:
                for fold in fcvs:
                    oof_count = t_count_dict.get(column_value, 0) - f_count_dict.get((column_value, fold), 0)
                    for target in tcvs:
                        oof_sum = t_sum_dict.get((column_value, target), 0) - col_agg_dict.get(
                            (column_value, fold, target), 0
                        )
                        alphas_values[(column_value, fold, target)] = [
                            (oof_sum + a * folds_prior_dict[(fold, target)]) / (oof_count + a) for a in self.alphas
                        ]

            def make_candidates(x):
                fold, target, column_value, count = x
                values = alphas_values[(column_value, fold, target)]
                for i, a in enumerate(self.alphas):
                    x[f"alpha_{i}"] = values[i]
                return x

            candidates_df = col_agg.apply(make_candidates, axis=1)

            best_alpha_index = np.array(
                [
                    (-np.log(candidates_df[f"alpha_{i}"]) * candidates_df["count"]).sum()
                    for i, a in enumerate(self.alphas)
                ]
            ).argmin()

            # bacn = f"alpha_{best_alpha_index}"
            # processing_df = pd.DataFrame(
            #     [[fv, tv, cv, alp[best_alpha_index]] for (cv, fv, tv), alp in alphas_values.items()],
            #     columns=[fcn, tcn, ccn, bacn]
            # )

            # mapping = processing_df.groupby([fcn, tcn, ccn]).max().to_dict()[bacn]
            # values = sc.broadcast(mapping)

            # for tcv in tcvs:
            #     cols_to_select.append(mcte_mapping_udf(values)(_fc, F.lit(tcv), _cc).alias(f"{self._fname_prefix}_{tcv}__{ccn}"))

            column_encodings_dict = (
                pd.DataFrame(
                    [
                        [
                            ccv,
                            tcv,
                            (t_sum_dict.get((ccv, tcv), 0) + self.alphas[best_alpha_index] * prior[tcv])
                            / (t_count_dict[ccv] + self.alphas[best_alpha_index]),
                        ]
                        for (ccv, fcv, tcv), _ in alphas_values.items()
                    ],
                    columns=[ccn, tcn, "encoding"],
                )
                .groupby([tcn, ccn])
                .max()
                .to_dict()["encoding"]
            )

            self.encodings.append(column_encodings_dict)

        logger.info(f"[{type(self)} (MCTE)] fit_transform is finished")

        return SparkMultiTargetEncoderTransformer(
            encodings=self.encodings,
            input_cols=self.getInputCols(),
            input_roles=self.getInputRoles(),
            output_cols=self.getOutputCols(),
            output_roles=self.getOutputRoles(),
            do_replace_columns=self.getDoReplaceColumns(),
        )


class SparkMultiTargetEncoderTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """
    Spark multiclass target encoder transformer.
    """

    _fit_checks = (categorical_check, multiclass_task_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "multioof"

    def __init__(
        self,
        encodings: List[Dict[str, Any]],
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool = False,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)
        self._encodings = encodings

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:

        cols_to_select = []
        logger.info(f"[{type(self)} (MCTE)] transform is started")

        sc = dataset.sql_ctx.sparkSession.sparkContext

        for i, (col_name, out_name) in enumerate(zip(self.getInputCols(), self.getOutputCols())):
            _cc = F.col(col_name)
            logger.debug(f"[{type(self)} (MCTE)] transform map size for column {col_name}: {len(self._encodings[i])}")

            enc = self._encodings[i]
            values = sc.broadcast(enc)
            for tcv in {tcv for tcv, _ in enc.keys()}:
                cols_to_select.append(mcte_transform_udf(values)(F.lit(tcv), _cc).alias(out_name))

        output = self._make_output_df(dataset, cols_to_select)

        logger.info(f"[{type(self)} (MCTE)] transform is finished")

        return output



import json
import logging
import os
import pickle
import time

from pathlib import Path
from typing import Union
from collections import namedtuple

from pyspark import SparkContext
from pyspark.sql.session import SparkSession
from pyspark.ml.common import inherit_doc
from pyspark.ml.util import DefaultParamsReader
from pyspark.ml.util import MLReadable
from pyspark.ml.util import MLReader
from pyspark.ml.util import MLWritable
from pyspark.ml.util import MLWriter
from synapse.ml.lightgbm import LightGBMClassificationModel
from synapse.ml.lightgbm import LightGBMRegressionModel
from synapse.ml.onnx import ONNXModel


logger = logging.getLogger(__name__)


class CommonPickleMLWritable(MLWritable):
    def write(self) -> MLWriter:
        "Returns MLWriter instance that can save the Transformer instance."
        return СommonPickleMLWriter(self)


class CommonPickleMLReadable(MLReadable):
    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return СommonPickleMLReader()


class СommonPickleMLWriter(MLWriter):
    """Implements saving an Estimator/Transformer instance to disk.
    Used when saving a trained pipeline.
    Implements MLWriter.saveImpl(path) method.
    """

    def __init__(self, instance):
        super().__init__()
        self.instance = instance

    def saveImpl(self, path: str) -> None:
        logger.info(f"Save {self.instance.__class__.__name__} to '{path}'")

        СommonPickleMLWriter.saveMetadata(self.instance, path, self.sc)

        pickled_instance = pickle.dumps(self.instance)
        Record = namedtuple("Record", ["pipeline"])
        rdd = self.sc.parallelize([Record(pickled_instance)])
        instance_df = rdd.map(lambda rec: Record(bytearray(rec.pipeline))).toDF()
        instance_df.write.mode("overwrite").parquet(os.path.join(path, "transformer_class_instance"))

    @staticmethod
    def saveMetadata(instance, path, sc):
        """
        Saves metadata + Params to: path + "/metadata"

        - class
        - timestamp
        - sparkVersion
        - uid
        - paramMap
        - defaultParamMap (since 2.4.0)
        - (optionally, extra metadata)

        Parameters
        ----------
        extraMetadata : dict, optional
            Extra metadata to be saved at same level as uid, paramMap, etc.
        paramMap : dict, optional
            If given, this is saved in the "paramMap" field.
        """
        metadataPath = os.path.join(path, "metadata")
        metadataJson = СommonPickleMLWriter._get_metadata_to_save(instance, sc)
        sc.parallelize([metadataJson], 1).saveAsTextFile(metadataPath)

    @staticmethod
    def _get_metadata_to_save(instance, sc):
        """
        Helper for :py:meth:`СommonPickleMLWriter.saveMetadata` which extracts the JSON to save.
        This is useful for ensemble models which need to save metadata for many sub-models.

        Notes
        -----
        See :py:meth:`DefaultParamsWriter.saveMetadata` for details on what this includes.
        """
        uid = instance.uid
        cls = instance.__module__ + "." + instance.__class__.__name__

        basicMetadata = {
            "class": cls,
            "timestamp": int(round(time.time() * 1000)),
            "sparkVersion": sc.version,
            "uid": uid,
            "paramMap": None,
            "defaultParamMap": None,
        }

        return json.dumps(basicMetadata, separators=[",", ":"])


class СommonPickleMLReader(MLReader):
    def load(self, path):
        """Load the ML instance from the input path."""

        spark_sess = SparkSession.builder.getOrCreate()
        df = spark_sess.read.parquet(os.path.join(path, "transformer_class_instance"))
        pickled_instance = df.rdd.map(lambda row: bytes(row.pipeline)).first()
        instance = pickle.loads(pickled_instance)

        return instance


class SparkLabelEncoderTransformerMLWritable(MLWritable):
    def write(self) -> MLWriter:
        "Returns MLWriter instance that can save the Transformer instance."
        return SparkLabelEncoderTransformerMLWriter(self)


class SparkLabelEncoderTransformerMLReadable(MLReadable):
    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return SparkLabelEncoderTransformerMLReader()


class SparkLabelEncoderTransformerMLWriter(MLWriter):
    """Implements saving an Estimator/Transformer instance to disk.
    Used when saving a trained pipeline.
    Implements MLWriter.saveImpl(path) method.
    """

    def __init__(self, instance):
        super().__init__()
        self.instance = instance

    def saveImpl(self, path: str) -> None:
        logger.info(f"Save {self.instance.__class__.__name__} to '{path}'")

        СommonPickleMLWriter.saveMetadata(self.instance, path, self.sc)

        indexer_model = self.instance.indexer_model
        self.instance.indexer_model = None
        pickled_instance = pickle.dumps(self.instance)
        self.instance.indexer_model = indexer_model
        Record = namedtuple("Record", ["pipeline"])
        rdd = self.sc.parallelize([Record(pickled_instance)])
        instance_df = rdd.map(lambda rec: Record(bytearray(rec.pipeline))).toDF()
        instance_df.write.mode("overwrite").parquet(os.path.join(path, "transformer_class_instance"))

        self.instance.indexer_model.write().overwrite().save(os.path.join(path, "indexer_model"))


class SparkLabelEncoderTransformerMLReader(MLReader):
    def load(self, path):
        """Load the ML instance from the input path."""

        spark_sess = SparkSession.builder.getOrCreate()
        df = spark_sess.read.parquet(os.path.join(path, "transformer_class_instance"))
        pickled_instance = df.rdd.map(lambda row: bytes(row.pipeline)).first()
        instance = pickle.loads(pickled_instance)

        from sparklightautoml.transformers.scala_wrappers.laml_string_indexer import LAMLStringIndexerModel

        indexer_model = LAMLStringIndexerModel.load(os.path.join(path, "indexer_model"))
        instance.indexer_model = indexer_model

        return instance


class ONNXModelWrapperMLWriter(MLWriter):
    def __init__(self, instance):
        super().__init__()
        self.instance = instance

    def saveImpl(self, path: str) -> None:
        logger.info(f"Save {self.instance.__class__.__name__} to '{path}'")

        СommonPickleMLWriter.saveMetadata(self.instance, path, self.sc)

        model: ONNXModel = self.instance.model
        model.save(os.path.join(path, "model"))


class ONNXModelWrapperMLReader(MLWriter):
    def load(self, path):
        """Load the ML instance from the input path and wrap by ONNXModelWrapper()"""

        from sparklightautoml.ml_algo.boost_lgbm import ONNXModelWrapper

        model_wrapper = ONNXModelWrapper()
        model_wrapper.model = ONNXModel.load(os.path.join(path, "model"))

        return model_wrapper


class LightGBMModelWrapperMLWriter(MLWriter):
    def __init__(self, instance):
        super().__init__()
        self.instance = instance

    def saveImpl(self, path: str) -> None:
        logger.info(f"Save {self.instance.__class__.__name__} to '{path}'")

        LightGBMModelWrapperMLWriter.saveMetadata(self.instance, path, self.sc)

        model: Union[LightGBMRegressionModel, LightGBMClassificationModel] = self.instance.model
        model.save(os.path.join(path, "model"))

    @staticmethod
    def saveMetadata(instance, path, sc):
        """
        Saves metadata + Params to: path + "/metadata"

        - class
        - timestamp
        - sparkVersion
        - uid
        - paramMap
        - defaultParamMap (since 2.4.0)
        - (optionally, extra metadata)

        Parameters
        ----------
        extraMetadata : dict, optional
            Extra metadata to be saved at same level as uid, paramMap, etc.
        paramMap : dict, optional
            If given, this is saved in the "paramMap" field.
        """
        metadataPath = os.path.join(path, "metadata")
        metadataJson = LightGBMModelWrapperMLWriter._get_metadata_to_save(instance, sc)
        sc.parallelize([metadataJson], 1).saveAsTextFile(metadataPath)

    @staticmethod
    def _get_metadata_to_save(instance, sc):
        """
        Helper for :py:meth:`LightGBMModelWrapperMLWriter.saveMetadata` which extracts the JSON to save.
        This is useful for ensemble models which need to save metadata for many sub-models.

        Notes
        -----
        See :py:meth:`LightGBMModelWrapperMLWriter.saveMetadata` for details on what this includes.
        """
        uid = instance.uid
        cls = instance.__module__ + "." + instance.__class__.__name__
        model_cls = instance.model.__module__ + "." + instance.model.__class__.__name__

        basicMetadata = {
            "class": cls,
            "timestamp": int(round(time.time() * 1000)),
            "sparkVersion": sc.version,
            "uid": uid,
            "paramMap": None,
            "defaultParamMap": None,
            "modelClass": model_cls,
        }

        return json.dumps(basicMetadata, separators=[",", ":"])


class LightGBMModelWrapperMLReader(MLReader):
    def load(self, path):
        """Load the ML instance from the input path and wrap by LightGBMModelWrapper()"""

        metadata = DefaultParamsReader.loadMetadata(path, self.sc)
        if metadata["modelClass"].endswith("LightGBMRegressionModel"):
            from synapse.ml.lightgbm.LightGBMRegressionModel import LightGBMRegressionModel as model_class
        elif metadata["modelClass"].endswith("LightGBMClassificationModel"):
            from synapse.ml.lightgbm.LightGBMClassificationModel import LightGBMClassificationModel as model_class
        else:
            raise NotImplementedError("Unknown model type.")

        from sparklightautoml.ml_algo.boost_lgbm import LightGBMModelWrapper

        model_wrapper = LightGBMModelWrapper()
        model_wrapper.model = model_class.load(os.path.join(path, "model"))

        return model_wrapper


@inherit_doc
class LAMLStringIndexerModelJavaMLReadable(MLReadable):
    """
    (Private) Mixin for instances that provide JavaMLReader.
    """

    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return LAMLStringIndexerModelJavaMLReader(cls)


def _jvm():
    """
    Returns the JVM view associated with SparkContext. Must be called
    after SparkContext is initialized.
    """
    jvm = SparkContext._jvm
    if jvm:
        return jvm
    else:
        raise AttributeError("Cannot load _jvm from SparkContext. Is SparkContext initialized?")


@inherit_doc
class LAMLStringIndexerModelJavaMLReader(MLReader):
    """
    (Private) Specialization of :py:class:`MLReader` for :py:class:`JavaParams` types
    """

    def __init__(self, clazz):
        super(LAMLStringIndexerModelJavaMLReader, self).__init__()
        self._clazz = clazz
        self._jread = self._load_java_obj(clazz).read()

    def load(self, path):
        """Load the ML instance from the input path."""
        if not isinstance(path, str):
            raise TypeError("path should be a string, got type %s" % type(path))
        java_obj = self._jread.load(path)
        if not hasattr(self._clazz, "_from_java"):
            raise NotImplementedError("This Java ML type cannot be loaded into Python currently: %r" % self._clazz)
        return self._clazz._from_java(java_obj)

    def session(self, sparkSession):
        """Sets the Spark Session to use for loading."""
        self._jread.session(sparkSession._jsparkSession)
        return self

    @classmethod
    def _load_java_obj(cls, clazz):
        """Load the peer Java object of the ML instance."""
        java_class = "org.apache.spark.ml.feature.lightautoml.LAMLStringIndexerModel"
        java_obj = _jvm()
        for name in java_class.split("."):
            java_obj = getattr(java_obj, name)
        return java_obj



from array import ArrayType
from copy import deepcopy
from typing import List, Optional, Dict, Tuple, Callable, Sequence, TypeVar, Iterator

import torch
import numpy as np
import pandas as pd
from pyspark import Broadcast
from pyspark.sql.pandas.functions import pandas_udf, PandasUDFType
from pyspark.sql.types import FloatType

from lightautoml.dataset.base import LAMLDataset
from lightautoml.image.image import DeepImageEmbedder, CreateImageFeatures
from lightautoml.image.utils import pil_loader
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.dataset.roles import NumericVectorOrArrayRole
from sparklightautoml.transformers.base import ObsoleteSparkTransformer
from lightautoml.text.utils import single_text_hash
from lightautoml.transformers.image import path_check


def vector_or_array_check(dataset: LAMLDataset):
    roles = dataset.roles
    features = dataset.features
    for f in features:
        assert isinstance(roles[f], NumericVectorOrArrayRole), "Only NumericVectorRole is accepted"


def path_or_vector_check(dataset: LAMLDataset):
    try:
        path_check(dataset)
        return
    except AssertionError:
        pass

    try:
        vector_or_array_check(dataset)
        return
    except AssertionError:
        pass

    assert False, "All incoming features should have same roles of either Path or NumericVector"


# TODO: needs a vector based alternative
class ImageFeaturesTransformer(ObsoleteSparkTransformer):
    """Simple image histogram."""

    _fit_checks = (path_check,)
    _transform_checks = ()
    _fname_prefix = "img_hist"

    _can_unwind_parents = False

    def __init__(
        self,
        hist_size: int = 30,
        is_hsv: bool = True,
        n_jobs: int = 4,
        loader: Callable = pil_loader,
    ):
        """Create normalized color histogram for rgb or hsv image.

        Args:
            hist_size: Number of bins for each channel.
            is_hsv: Convert image to hsv.
            n_jobs: Number of threads for multiprocessing.
            loader: Callable for reading image from path.

        """
        self.hist_size = hist_size
        self.is_hsv = is_hsv
        self.n_jobs = n_jobs
        self.loader = loader
        self._fg: Optional[CreateImageFeatures] = None

    @property
    def features(self) -> List[str]:
        """Features list.

        Returns:
            List of features names.

        """
        return self._features

    def _fit(self, dataset: SparkDataset):
        """Init hist class and create feature names.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        Returns:
            self.

        """

        sdf = dataset.data
        self._fg = CreateImageFeatures(self.hist_size, self.is_hsv, self.n_jobs, self.loader)
        self._features = [f"{self._fname_prefix}__{c}" for c in sdf.columns]

        return self

    def _transform(self, dataset: SparkDataset) -> SparkDataset:
        """Transform image dataset to color histograms.

        Args:
            dataset: Pandas or Numpy dataset of image paths.

        Returns:
            Dataset with encoded text.

        """

        sdf = dataset.data

        # transform
        roles = []
        new_cols = []
        for c, out_col_name in zip(dataset.features, self.features):
            role = NumericVectorOrArrayRole(
                size=len(self._fg.fe.get_names()),
                element_col_name_template=[f"{self._fname_prefix}_{name}__{c}" for name in self._fg.fe.get_names()],
                dtype=np.float32,
                is_vector=False,
            )
            fg_bcast = dataset.spark_session.sparkContext.broadcast(self._fg)

            @pandas_udf("array<float>", PandasUDFType.SCALAR)
            def calculate_embeddings(data: pd.Series) -> pd.Series:
                fg = fg_bcast.value
                img_embeds = pd.Series(list(fg.transform(data)))
                return img_embeds

            new_cols.append(calculate_embeddings(c).alias(out_col_name))
            roles.append(role)

        new_sdf = sdf.select(*dataset.service_columns, *new_cols)

        output = dataset.empty()
        output.set_data(new_sdf, self.features, roles)

        return output


class AutoCVWrap(ObsoleteSparkTransformer):
    """Calculate image embeddings."""

    _fit_checks = ()
    _transform_checks = ()
    _fname_prefix = "emb_cv"
    _emb_name = ""

    _T = TypeVar("_T")

    @property
    def features(self) -> List[str]:
        """Features list.

        Returns:
            List of features names.

        """
        return self._features

    @property
    def _image_loader(self):
        raise NotImplementedError()

    def __init__(
        self,
        model="efficientnet-b0",
        weights_path: Optional[str] = None,
        cache_dir: str = "./cache_CV",
        subs: Optional = None,
        device: torch.device = torch.device("cuda:0"),
        n_jobs: int = 4,
        random_state: int = 42,
        is_advprop: bool = True,
        batch_size: int = 128,
        verbose: bool = True,
    ):
        """

        Args:
            model: Name of effnet model.
            weights_path: Path to saved weights.
            cache_dir: Path to cache directory or None.
            subs: Subsample to fit transformer. If ``None`` - full data.
            device: Torch device.
            n_jobs: Number of threads for dataloader.
            random_state: Random state to take subsample and set torch seed.
            is_advprop: Use adversarial training.
            batch_size: Batch size for embedding model.
            verbose: Verbose data processing.

        """
        self.embed_model = model
        self.random_state = random_state
        self.subs = subs
        self.cache_dir = cache_dir
        self._img_transformers: Optional[Dict[str, Tuple[DeepImageEmbedder, str]]] = None

        self.transformer = DeepImageEmbedder(
            device,
            n_jobs,
            random_state,
            is_advprop,
            model,
            weights_path,
            batch_size,
            verbose,
            image_loader=self._image_loader,
        )

        self._emb_name = "DI_" + single_text_hash(self.embed_model)
        self.emb_size = self.transformer.model.feature_shape

    def _fit(self, dataset: SparkDataset):
        """Fit chosen transformer and create feature names.

        Args:
            dataset: Pandas or Numpy dataset of text features.

        """

        sdf = dataset.data

        self._img_transformers = dict()
        for c in dataset.features:
            out_column_name = f"{self._fname_prefix}_{self._emb_name}__{c}"

            self._img_transformers[c] = (
                # TODO: we don't really want 'fit' here, because it would happen on the driver side
                # TODO: better to mark fitless classes with some Marker type via inheritance
                # TODO: to avoid errors of applying the wrong transformer as early as possible
                deepcopy(self.transformer.fit(sdf.select(c))),
                out_column_name,
            )

        self._features = [feat for _, feat in self._img_transformers.values()]
        return self

    def _transform(self, dataset: SparkDataset) -> SparkDataset:
        """Transform dataset to image embeddings.

        Args:
            dataset: Pandas or Numpy dataset of image paths.

        Returns:
            Numpy dataset with image embeddings.

        """

        sdf = dataset.data

        # transform
        roles = []
        new_cols = []
        for c in dataset.features:
            role = NumericVectorOrArrayRole(
                size=self.emb_size,
                element_col_name_template=f"{self._fname_prefix}_{self._emb_name}_{{}}__{c}",
                dtype=np.float32,
                is_vector=False,
            )

            # TODO: probably transformer should be created on the worker side and not in the driver
            trans, out_col_name = self._img_transformers[c]
            transformer_bcast = dataset.spark_session.sparkContext.broadcast(value=trans)

            @pandas_udf("array<float>", PandasUDFType.SCALAR)
            def calculate_embeddings(data: pd.Series) -> pd.Series:
                transformer = transformer_bcast.value
                img_embeds = pd.Series(list(transformer.transform(data)))
                return img_embeds

            new_cols.append(calculate_embeddings(c).alias(out_col_name))
            roles.append(role)

        new_sdf = sdf.select(*dataset.service_columns, *new_cols)

        output = dataset.empty()
        output.set_data(new_sdf, self.features, roles)

        return output


class PathBasedAutoCVWrap(AutoCVWrap):
    _fit_checks = (path_check,)
    _T = str

    def __init__(self, image_loader: Callable, *args, **kwargs):
        self.__image_loader = image_loader
        super().__init__(*args, **kwargs)

    @property
    def _image_loader(self):
        return self.__image_loader


class ArrayBasedAutoCVWrap(AutoCVWrap):
    _fit_checks = (vector_or_array_check,)
    _T = bytes

    def __init__(self, *args, **kwargs):
        self.__image_loader = lambda x: x
        super().__init__(*args, **kwargs)

    @property
    def _image_loader(self):
        return self.__image_loader



import logging
from abc import ABC
from copy import copy
from typing import List, Optional, Sequence, Tuple, cast, Callable

import numpy as np
from pyspark.ml import Transformer
from pyspark.ml.param import Params
from pyspark.ml.param.shared import HasInputCols, HasOutputCol, Param
from pyspark.ml.util import MLWritable
from pyspark.sql import functions as F
from pyspark.sql.functions import isnan

from lightautoml.automl.blend import WeightedBlender
from lightautoml.dataset.np_pd_dataset import NumpyDataset
from lightautoml.dataset.roles import ColumnRole, NumericRole
from lightautoml.reader.base import RolesDict
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.utils import SparkDataFrame
from sparklightautoml.dataset.roles import NumericVectorOrArrayRole
from sparklightautoml.ml_algo.base import AveragingTransformer
from sparklightautoml.pipelines.ml.base import SparkMLPipeline
from sparklightautoml.tasks.base import DEFAULT_PREDICTION_COL_NAME, SparkTask
from sparklightautoml.transformers.base import ColumnsSelectorTransformer


logger = logging.getLogger(__name__)


class SparkBlender(ABC):
    """Basic class for blending.

    Blender learns how to make blend
    on sequence of prediction datasets and prune pipes,
    that are not used in final blend.

    """

    def __init__(self):
        self._transformer = None
        self._single_prediction_col_name = DEFAULT_PREDICTION_COL_NAME
        self._pred_role: Optional[ColumnRole] = None
        self._output_roles: Optional[RolesDict] = None
        self._task: Optional[SparkTask] = None

    @property
    def output_roles(self) -> RolesDict:
        assert self._output_roles is not None, "Blender has not been fitted yet"
        return self._output_roles

    @property
    def transformer(self) -> Transformer:
        """Returns Spark MLlib Transformer.
        Represents a Transformer with fitted models."""

        assert self._transformer is not None, "Pipeline is not fitted!"

        return self._transformer

    def fit_predict(
        self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]
    ) -> Tuple[SparkDataset, Sequence[SparkMLPipeline]]:
        logger.info(f"Blender {type(self)} starting fit_predict")

        if len(pipes) == 1 and len(pipes[0].ml_algos) == 1:
            self._transformer = ColumnsSelectorTransformer(
                input_cols=[SparkDataset.ID_COLUMN] + list(pipes[0].output_roles.keys()),
                optional_cols=[predictions.target_column] if predictions.target_column else [],
            )
            self._output_roles = copy(predictions.roles)
            return predictions, pipes

        self._set_metadata(predictions, pipes)

        result = self._fit_predict(predictions, pipes)

        logger.info(f"Blender {type(self)} finished fit_predict")

        return result

    def predict(self, predictions: SparkDataset) -> SparkDataset:
        sdf = self._transformer.transform(predictions.data)

        sds = predictions.empty()
        sds.set_data(sdf, list(self.output_roles.keys()), self.output_roles)

        return sds

    def _fit_predict(
        self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]
    ) -> Tuple[SparkDataset, Sequence[SparkMLPipeline]]:
        raise NotImplementedError()

    def split_models(self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]) -> List[Tuple[str, int, int]]:
        """Split predictions by single model prediction datasets.

        Args:
            predictions: Dataset with predictions.

        Returns:
            Each tuple in the list is:
            - prediction column name
            - corresponding model index (in the pipe)
            - corresponding pipe index

        """
        return [
            (ml_algo.prediction_feature, j, i)
            for i, pipe in enumerate(pipes)
            for j, ml_algo in enumerate(pipe.ml_algos)
        ]

    def _set_metadata(self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]):
        self._pred_role = predictions.roles[pipes[0].ml_algos[0].prediction_feature]
        self._output_roles = {self._single_prediction_col_name: self._pred_role}

        if isinstance(self._pred_role, NumericVectorOrArrayRole):
            self._outp_dim = self._pred_role.size
        else:
            self._outp_dim = 1
        self._outp_prob = predictions.task.name in ["binary", "multiclass"]
        self._score = predictions.task.get_dataset_metric()
        self._task = predictions.task

    def _make_single_pred_ds(self, predictions: SparkDataset, pred_col: str) -> SparkDataset:
        pred_sdf = predictions.data.select(
            SparkDataset.ID_COLUMN, predictions.target_column, F.col(pred_col).alias(self._single_prediction_col_name)
        )
        pred_roles = {c: predictions.roles[c] for c in pred_sdf.columns}
        pred_ds = predictions.empty()
        pred_ds.set_data(pred_sdf, pred_sdf.columns, pred_roles)

        return pred_ds

    def score(self, dataset: SparkDataset) -> float:
        """Score metric for blender.

        Args:
            dataset: Blended predictions dataset.

        Returns:
            Metric value.

        """
        return self._score(dataset, True)


class SparkBestModelSelector(SparkBlender, WeightedBlender):
    """Select best single model from level.

    Drops pipes that are not used in calc best model.
    Works in general case (even on some custom things)
    and most efficient on inference.
    Perform worse than other on tables,
    specially if some of models was terminated by timer.

    """

    def _fit_predict(
        self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]
    ) -> Tuple[SparkDataset, Sequence[SparkMLPipeline]]:
        """Simple fit - just take one best.

        Args:
            predictions: Sequence of datasets with predictions.
            pipes: Sequence of pipelines.

        Returns:
            Single prediction dataset and Sequence of pruned pipelines.

        """
        splitted_models_and_pipes = self.split_models(predictions, pipes)

        best_pred = None
        best_pipe_idx = 0
        best_model_idx = 0
        best_score = -np.inf

        for pred_col, mod, pipe in splitted_models_and_pipes:
            pred_ds = self._make_single_pred_ds(predictions, pred_col)
            score = self.score(pred_ds)

            if score > best_score:
                best_pipe_idx = pipe
                best_model_idx = mod
                best_score = score
                best_pred = pred_ds

        best_pipe = pipes[best_pipe_idx]
        best_pipe.ml_algos = [best_pipe.ml_algos[best_model_idx]]

        self._transformer = ColumnsSelectorTransformer(
            input_cols=[SparkDataset.ID_COLUMN, self._single_prediction_col_name]
        )

        self._output_roles = copy(best_pred.roles)

        return best_pred, [best_pipe]


class SparkWeightedBlender(SparkBlender, WeightedBlender):
    """Weighted Blender based on coord descent, optimize task metric directly.

    Weight sum eq. 1.
    Good blender for tabular data,
    even if some predictions are NaN (ex. timeout).
    Model with low weights will be pruned.

    """

    def __init__(
        self,
        max_iters: int = 5,
        max_inner_iters: int = 7,
        max_nonzero_coef: float = 0.05,
    ):
        SparkBlender.__init__(self)
        WeightedBlender.__init__(self, max_iters, max_inner_iters, max_nonzero_coef)
        self._predictions_dataset: Optional[SparkDataset] = None

    def _get_weighted_pred(
        self,
        splitted_preds: Sequence[str],
        wts: Optional[np.ndarray],
        remove_splitted_preds_cols: Optional[List[str]] = None,
    ) -> SparkDataset:
        avr = self._build_avr_transformer(splitted_preds, wts, remove_splitted_preds_cols)

        weighted_preds_sdf = avr.transform(self._predictions_dataset.data)

        wpreds_sds = self._predictions_dataset.empty()
        wpreds_sds.set_data(weighted_preds_sdf, list(self.output_roles.keys()), self.output_roles)

        return wpreds_sds

    def _build_avr_transformer(
        self,
        splitted_preds: Sequence[str],
        wts: Optional[np.ndarray],
        remove_splitted_preds_cols: Optional[List[str]] = None,
    ) -> AveragingTransformer:
        remove_cols = list(splitted_preds)

        if remove_splitted_preds_cols is not None:
            remove_cols.extend(remove_splitted_preds_cols)

        return AveragingTransformer(
            task_name=self._task.name,
            input_cols=list(splitted_preds),
            output_col=self._single_prediction_col_name,
            remove_cols=remove_cols,
            convert_to_array_first=True,
            weights=(wts * len(wts)).tolist(),
            dim_num=self._outp_dim,
        )

    def _fit_predict(
        self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]
    ) -> Tuple[SparkDataset, Sequence[SparkMLPipeline]]:

        self._predictions_dataset = predictions

        sm = self.split_models(predictions, pipes)
        pred_cols = [pred_col for pred_col, _, _ in sm]
        pipe_idx = np.array([pidx for _, _, pidx in sm])

        wts = self._optimize(pred_cols)

        reweighted_pred_cols = [x for (x, w) in zip(pred_cols, wts) if w > 0]
        removed_cols = [x for x in pred_cols if x not in reweighted_pred_cols]
        _, self.wts = self._prune_pipe(pipes, wts, pipe_idx)
        pipes = cast(Sequence[SparkMLPipeline], pipes)

        self._transformer = self._build_avr_transformer(
            reweighted_pred_cols, self.wts, remove_splitted_preds_cols=removed_cols
        )
        outp = self._get_weighted_pred(reweighted_pred_cols, self.wts, remove_splitted_preds_cols=removed_cols)

        return outp, pipes


class SparkMeanBlender(SparkBlender):
    """Simple average level predictions.

    Works only with TabularDatasets.
    Doesn't require target to fit.
    No pruning.

    """

    def _fit_predict(
        self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]
    ) -> Tuple[SparkDataset, Sequence[SparkMLPipeline]]:

        pred_cols = [pred_col for pred_col, _, _ in self.split_models(predictions, pipes)]

        self._transformer = AveragingTransformer(
            task_name=predictions.task.name,
            input_cols=pred_cols,
            output_col=self._single_prediction_col_name,
            remove_cols=pred_cols,
            convert_to_array_first=not (predictions.task.name == "reg"),
            dim_num=self._outp_dim,
        )

        df = self._transformer.transform(predictions.data)

        if predictions.task.name in ["binary", "multiclass"]:
            assert isinstance(self._pred_role, NumericVectorOrArrayRole)
            output_role = NumericVectorOrArrayRole(
                self._pred_role.size,
                f"MeanBlend_{{}}",
                dtype=np.float32,
                prob=self._outp_prob,
                is_vector=self._pred_role.is_vector,
            )
        else:
            output_role = NumericRole(np.float32, prob=self._outp_prob)

        roles = {f: predictions.roles[f] for f in predictions.features if f not in pred_cols}
        roles[self._single_prediction_col_name] = output_role
        pred_ds = predictions.empty()
        pred_ds.set_data(df, df.columns, roles)

        self._output_roles = copy(roles)

        return pred_ds, pipes



"""Base class for selection pipelines."""
from copy import copy
from typing import Any, Optional, List

from lightautoml.dataset.base import LAMLDataset
from lightautoml.pipelines.selection.base import SelectionPipeline
from lightautoml.validation.base import TrainValidIterator
from pandas import Series

from sparklightautoml.dataset.base import SparkDataset


class SparkImportanceEstimator:
    """
    Abstract class, that estimates feature importances.
    """

    def __init__(self):
        self.raw_importances = None

    # Change signature here to be compatible with MLAlgo
    def fit(self, *args: Any, **kwargs: Any):
        raise NotImplementedError

    def get_features_score(self) -> SparkDataset:

        return self.raw_importances


class BugFixSelectionPipelineWrapper(SelectionPipeline):
    def __init__(self, instance: SelectionPipeline):
        super(BugFixSelectionPipelineWrapper, self).__init__()
        self._instance = instance

    def fit(self, train_valid: TrainValidIterator):
        self._instance.fit(train_valid)

    def map_raw_feature_importances(self, raw_importances: Series):
        return self._instance.map_raw_feature_importances(raw_importances)

    def get_features_score(self):
        return self._instance.get_features_score()

    @property
    def is_fitted(self) -> bool:
        return self._instance.is_fitted

    @property
    def selected_features(self) -> List[str]:
        return self._instance.selected_features

    @property
    def in_features(self) -> List[str]:
        return self._instance.in_features

    @property
    def dropped_features(self) -> List[str]:
        return self._instance.dropped_features

    def perform_selection(self, train_valid: Optional[TrainValidIterator]):
        self._instance.perform_selection(train_valid)

    def select(self, dataset: LAMLDataset) -> LAMLDataset:
        selected_features = copy(self._instance.selected_features)
        # Add features that forces input
        sl_set = set(selected_features)
        roles = dataset.roles
        for col in (x for x in dataset.features if x not in sl_set):
            if roles[col].force_input:
                if col not in sl_set:
                    selected_features.append(col)

        return dataset[:, selected_features]



from pyspark import keyword_only
from pyspark.ml.common import inherit_doc
from pyspark.ml.wrapper import JavaTransformer


@inherit_doc
class BalancedUnionPartitionsCoalescerTransformer(JavaTransformer):
    """
    Custom implementation of PySpark BalancedUnionPartitionsCoalescerTransformer wrapper
    """

    @keyword_only
    def __init__(self):
        super(BalancedUnionPartitionsCoalescerTransformer, self).__init__()
        self._java_obj = self._new_java_obj(
            "org.apache.spark.lightautoml.utils.BalancedUnionPartitionsCoalescerTransformer", self.uid
        )



import logging
import os
import shutil
from typing import Optional, Any, Sequence, Iterable

import torch
import yaml

from sparklightautoml.automl.base import SparkAutoML
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.tasks.base import SparkTask
from lightautoml.utils.logging import verbosity_to_loglevel, set_stdout_level, add_filehandler
from lightautoml.utils.timer import PipelineTimer


logger = logging.getLogger(__name__)

base_dir = os.path.dirname(__file__)


class SparkAutoMLPreset(SparkAutoML):
    """Basic class for automl preset.

    It's almost like AutoML, but with delayed initialization.
    Initialization starts on fit, some params are inferred from data.
    Preset should be defined via ``.create_automl`` method.
    Params should be set via yaml config.
    Most usefull case - end-to-end model development.

    Example:

        >>> automl = SomePreset(SparkTask('binary'), timeout=3600)
        >>> automl.fit_predict(data, roles={'target': 'TARGET'})

    """

    _default_config_path = "example_config.yml"

    def __init__(
        self,
        task: SparkTask,
        timeout: int = 3600,
        memory_limit: int = 16,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
        timing_params: Optional[dict] = None,
        config_path: Optional[str] = None,
        **kwargs: Any,
    ):
        """

        Commonly _params kwargs (ex. timing_params) set via
        config file (config_path argument).
        If you need to change just few params,
        it's possible to pass it as dict of dicts, like json.
        To get available params please look on default config template.
        Also you can find there param description.
        To generate config template
        call ``SomePreset.get_config('config_path.yml')``.

        Args:
            task: Task to solve.
            timeout: Timeout in seconds.
            memory_limit: Memory limit that are passed to each automl.
            cpu_limit: CPU limit that that are passed to each automl.
            gpu_ids: GPU IDs that are passed to each automl.
            verbose: Controls the verbosity: the higher, the more messages.
                <1  : messages are not displayed;
                >=1 : the computation process for layers is displayed;
                >=2 : the information about folds processing is also displayed;
                >=3 : the hyperparameters optimization process is also displayed;
                >=4 : the training process for every algorithm is displayed;
            timing_params: Timing param dict.
            config_path: Path to config file.
            **kwargs: Not used.

        """
        self._set_config(config_path)

        for name, param in zip(["timing_params"], [timing_params]):
            if param is None:
                param = {}
            self.__dict__[name] = {**self.__dict__[name], **param}

        self.timer = PipelineTimer(timeout, **getattr(self, "timing_params"))
        self.memory_limit = memory_limit
        if cpu_limit == -1:
            cpu_limit = os.cpu_count()
        self.cpu_limit = cpu_limit
        self.gpu_ids = gpu_ids
        if gpu_ids == "all":
            self.gpu_ids = ",".join(map(str, range(torch.cuda.device_count())))
        self.task = task

    def _set_config(self, path):
        self.config_path = path

        if path is None:
            path = os.path.join(base_dir, self._default_config_path)

        with open(path) as f:
            params = yaml.safe_load(f)

        for k in params:
            self.__dict__[k] = params[k]

    @classmethod
    def get_config(cls, path: Optional[str] = None) -> Optional[dict]:
        """Create new config template.

        Args:
            path: Path to config.

        Returns:
            Config.

        """
        if path is None:
            path = os.path.join(base_dir, cls._default_config_path)
            with open(path) as f:
                params = yaml.safe_load(f)
            return params

        else:
            shutil.copy(os.path.join(base_dir, cls._default_config_path), path)

    def create_automl(self, **fit_args):
        """Abstract method - how to build automl.

        Here you should create all automl components,
        like readers, levels, timers, blenders.
        Method ``._initialize`` should be called in the end to create automl.

        Args:
            **fit_args: params that are passed to ``.fit_predict`` method.

        """
        raise NotImplementedError

    def fit_predict(
        self,
        train_data: Any,
        roles: dict,
        train_features: Optional[Sequence[str]] = None,
        cv_iter: Optional[Iterable] = None,
        valid_data: Optional[Any] = None,
        valid_features: Optional[Sequence[str]] = None,
        verbose: int = 0,
    ) -> SparkDataset:
        """Fit on input data and make prediction on validation part.

        Args:
            train_data: Dataset to train.
            roles: Roles dict.
            train_features: Features names,
              if can't be inferred from `train_data`.
            cv_iter: Custom cv-iterator. For example,
              :class:`~lightautoml.validation.np_iterators.TimeSeriesIterator`.
            valid_data: Optional validation dataset.
            valid_features: Optional validation dataset features if can't be
              inferred from `valid_data`.
            verbose: Verbosity level that are passed to each automl.

        Returns:
            Dataset with predictions. Call ``.data`` to get predictions array.

        """
        self.set_verbosity_level(verbose)

        self.create_automl(
            train_data=train_data,
            roles=roles,
            train_features=train_features,
            cv_iter=cv_iter,
            valid_data=valid_data,
            valid_features=valid_features,
        )
        logger.info(f"Task: {self.task.name}\n")

        logger.info("Start automl preset with listed constraints:")
        logger.info(f"- time: {self.timer.timeout:.2f} seconds")
        logger.info(f"- CPU: {self.cpu_limit} cores")
        logger.info(f"- memory: {self.memory_limit} GB\n")

        self.timer.start()
        result = super().fit_predict(
            train_data,
            roles,
            train_features,
            cv_iter,
            valid_data,
            valid_features,
            verbose,
        )

        logger.info("\x1b[1mAutoml preset training completed in {:.2f} seconds\x1b[0m\n".format(self.timer.time_spent))
        logger.info(f"Model description:\n{self.create_model_str_desc()}\n")

        return result

    def create_model_str_desc(self, pref_tab_num: int = 0, split_line_len: int = 0) -> str:
        prefix = "\t" * pref_tab_num
        splitter = prefix + "=" * split_line_len + "\n"
        model_stats = sorted(list(self.collect_model_stats().items()))

        last_lvl = model_stats[-1][0].split("_")[1]
        last_lvl_models = [ms for ms in model_stats if ms[0].startswith("Lvl_" + last_lvl)]
        notlast_lvl_models = [ms for ms in model_stats if not ms[0].startswith("Lvl_" + last_lvl)]

        res = ""
        if len(notlast_lvl_models) > 0:
            cur_level = 0
            res += prefix + "Models on level 0:\n"
            for model_stat in notlast_lvl_models:
                model_name, cnt_folds = model_stat
                level = int(model_name.split("_")[1])
                if level != cur_level:
                    cur_level = level
                    res += "\n" + prefix + "Models on level {}:\n".format(cur_level)
                res += prefix + "\t {} averaged models {}\n".format(cnt_folds, model_name)
            res += "\n"

        res += prefix + "Final prediction for new objects (level {}) = \n".format(last_lvl)
        for model_stat, weight in zip(last_lvl_models, self.blender.wts):
            model_name, cnt_folds = model_stat
            res += prefix + "\t {:.5f} * ({} averaged models {}) +\n".format(weight, cnt_folds, model_name)

        if split_line_len == 0:
            return res[:-2]

        return splitter + res[:-2] + "\n" + splitter

    @staticmethod
    def set_verbosity_level(verbose: int):
        """Verbosity level setter.

        Args:
            verbose: Controls the verbosity: the higher, the more messages.
                <1  : messages are not displayed;
                >=1 : the computation process for layers is displayed;
                >=2 : the information about folds processing is also displayed;
                >=3 : the hyperparameters optimization process is also displayed;
                >=4 : the training process for every algorithm is displayed;

        """
        level = verbosity_to_loglevel(verbose)
        set_stdout_level(level)

        logger.info(f"Stdout logging level is {logging._levelToName[level]}.")

    @staticmethod
    def set_logfile(filename: str):
        """"""
        add_filehandler(filename)



"""Report generators and templates."""

# from lightautoml.utils.installation import __validate_extra_deps

from .report_deco import SparkReportDeco


# __validate_extra_deps("pdf")


__all__ = ["SparkReportDeco"]



import logging
import socket
import time
import warnings
from contextlib import contextmanager
from datetime import datetime
from typing import Optional, Tuple, Dict

import pyspark
from pyspark import RDD
from pyspark.ml import Transformer, Estimator
from pyspark.ml.util import DefaultParamsWritable, DefaultParamsReadable
from pyspark.sql import SparkSession

VERBOSE_LOGGING_FORMAT = "%(asctime)s %(levelname)s %(module)s %(filename)s:%(lineno)d %(message)s"

logger = logging.getLogger(__name__)


@contextmanager
def spark_session(
    session_args: Optional[dict] = None, master: str = "local[]", wait_secs_after_the_end: Optional[int] = None
) -> SparkSession:
    """
    Args:
        master: address of the master
            to run locally - "local[1]"

            to run on spark cluster - "spark://node4.bdcl:7077"
            (Optionally set the driver host to a correct hostname .config("spark.driver.host", "node4.bdcl"))

        wait_secs_after_the_end: amount of seconds to wait before stoping SparkSession and thus web UI.

    Returns:
        SparkSession to be used and that is stopped upon exiting this context manager
    """

    if not session_args:
        spark_sess_builder = (
            SparkSession.builder.appName("SPARK-LAMA-app")
            .master(master)
            .config("spark.jars", "jars/spark-lightautoml_2.12-0.1.jar")
            .config("spark.jars.packages", "com.microsoft.azure:synapseml_2.12:0.9.5")
            .config("spark.jars.repositories", "https://mmlspark.azureedge.net/maven")
            .config("spark.sql.shuffle.partitions", "16")
            .config("spark.kryoserializer.buffer.max", "512m")
            .config("spark.driver.cores", "4")
            .config("spark.driver.memory", "16g")
            .config("spark.cores.max", "16")
            .config("spark.executor.instances", "4")
            .config("spark.executor.memory", "16g")
            .config("spark.executor.cores", "4")
            .config("spark.memory.fraction", "0.6")
            .config("spark.memory.storageFraction", "0.5")
            .config("spark.sql.autoBroadcastJoinThreshold", "100MB")
            .config("spark.sql.execution.arrow.pyspark.enabled", "true")
        )
    else:
        spark_sess_builder = SparkSession.builder.appName("SPARK-LAMA-app")
        for arg, value in session_args.items():
            spark_sess_builder = spark_sess_builder.config(arg, value)

        if "spark.master" in session_args and not session_args["spark.master"].startswith("local["):
            local_ip_address = socket.gethostbyname(socket.gethostname())
            logger.info(f"Using IP address for spark driver: {local_ip_address}")
            spark_sess_builder = spark_sess_builder.config("spark.driver.host", local_ip_address)

    spark_sess = spark_sess_builder.getOrCreate()

    logger.info(f"Spark WebUI url: {spark_sess.sparkContext.uiWebUrl}")

    try:
        yield spark_sess
    finally:
        logger.info(
            f"The session is ended. Sleeping {wait_secs_after_the_end if wait_secs_after_the_end else 0} "
            f"secs until stop the spark session."
        )
        if wait_secs_after_the_end:
            time.sleep(wait_secs_after_the_end)
        spark_sess.stop()


@contextmanager
def log_exec_time(name: Optional[str] = None, write_log=True):

    # Add file handler for INFO
    if write_log:
        file_handler_info = logging.FileHandler(f"/tmp/{name}_log.log.log", mode="a")
        file_handler_info.setFormatter(logging.Formatter("%(message)s"))
        file_handler_info.setLevel(logging.INFO)
        logger.addHandler(file_handler_info)

    start = datetime.now()

    yield

    end = datetime.now()
    duration = (end - start).total_seconds()

    msg = f"Exec time of {name}: {duration}" if name else f"Exec time: {duration}"
    logger.warning(msg)


# log_exec_time() class to return elapsed time value
class log_exec_timer:
    def __init__(self, name: Optional[str] = None):
        self.name = name
        self._start = None
        self._duration = None

    def __enter__(self):
        self._start = datetime.now()
        return self

    def __exit__(self, type, value, traceback):
        self._duration = (datetime.now() - self._start).total_seconds()
        msg = f"Exec time of {self.name}: {self._duration}" if self.name else f"Exec time: {self._duration}"
        logger.info(msg)

    @property
    def duration(self):
        return self._duration


SparkDataFrame = pyspark.sql.DataFrame


def get_cached_df_through_rdd(df: SparkDataFrame, name: Optional[str] = None) -> Tuple[SparkDataFrame, RDD]:
    rdd = df.rdd
    cached_rdd = rdd.setName(name).cache() if name else rdd.cache()
    cached_df = df.sql_ctx.createDataFrame(cached_rdd, df.schema)
    return cached_df, cached_rdd


def logging_config(level: int = logging.INFO, log_filename: str = "/var/log/lama.log") -> dict:
    return {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "verbose": {"format": VERBOSE_LOGGING_FORMAT},
            "simple": {"format": "%(asctime)s %(levelname)s %(message)s"},
        },
        "handlers": {
            "console": {"level": "DEBUG", "class": "logging.StreamHandler", "formatter": "verbose"},
            "file": {
                "level": "DEBUG",
                "class": "logging.FileHandler",
                "formatter": "verbose",
                "filename": log_filename,
            },
        },
        "loggers": {
            "lightautoml": {
                "handlers": ["console", "file"],
                "propagate": True,
                "level": level,
            },
            "sparklightautoml": {
                "handlers": ["console", "file"],
                "level": level,
                "propagate": False,
            },
            "lightautoml.ml_algo": {
                "handlers": ["console", "file"],
                "level": level,
                "propagate": False,
            },
        },
    }


def cache(df: SparkDataFrame) -> SparkDataFrame:
    if not df.is_cached:
        df = df.cache()
    return df


def warn_if_not_cached(df: SparkDataFrame):
    if not df.is_cached:
        warnings.warn(
            "Attempting to calculate shape on not cached dataframe. " "It may take too much time.", RuntimeWarning
        )


class NoOpTransformer(Transformer, DefaultParamsWritable, DefaultParamsReadable):
    def __init__(self, name: Optional[str] = None):
        super().__init__()
        self._name = name

    def _transform(self, dataset):
        return dataset


class DebugTransformer(Transformer):
    def __init__(self, name: Optional[str] = None):
        super().__init__()
        self._name = name

    def _transform(self, dataset):
        dataset = dataset.cache()
        dataset.write.mode("overwrite").format("noop").save()
        return dataset


class Cacher(Estimator):
    _cacher_dict: Dict[str, SparkDataFrame] = dict()

    @classmethod
    def get_dataset_by_key(cls, key: str) -> Optional[SparkDataFrame]:
        return cls._cacher_dict.get(key, None)

    @classmethod
    def release_cache_by_key(cls, key: str):
        df = cls._cacher_dict.pop(key, None)
        if df is not None:
            df.unpersist()
            del df

    @property
    def dataset(self) -> SparkDataFrame:
        """Returns chached dataframe"""
        return self._cacher_dict[self._key]

    def __init__(self, key: str):
        super().__init__()
        self._key = key
        self._dataset: Optional[SparkDataFrame] = None

    def _fit(self, dataset):
        logger.info(f"Cacher {self._key} (RDD Id: {dataset.rdd.id()}). Starting to materialize data.")

        # using local checkpoints
        ds = dataset.localCheckpoint(eager=True)

        # # using plain caching
        # ds = SparkSession.getActiveSession().createDataFrame(dataset.rdd, schema=dataset.schema).cache()
        # ds.write.mode('overwrite').format('noop').save()

        logger.info(f"Cacher {self._key} (RDD Id: {ds.rdd.id()}). Finished data materialization.")

        previous_ds = self._cacher_dict.get(self._key, None)
        if previous_ds is not None and ds != previous_ds:
            logger.info(f"Removing cache for key: {self._key} (RDD Id: {previous_ds.rdd.id()}).")
            previous_ds.unpersist()
            del previous_ds

        self._cacher_dict[self._key] = ds

        return NoOpTransformer(name=f"cacher_{self._key}")


class EmptyCacher(Cacher):
    def __init__(self, key: str):
        super().__init__(key)
        self._dataset: Optional[SparkDataFrame] = None

    @property
    def dataset(self) -> SparkDataFrame:
        return self._dataset

    def _fit(self, dataset):
        self._dataset = dataset
        return NoOpTransformer(name=f"empty_cacher_{self._key}")
